import copy
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    scope: str,
    default_action: dict,
    visibility_config: dict,
    resource_id: str = None,
    description: str = None,
    rules: list = None,
    custom_response_bodies: dict = None,
    captcha_config: dict = None,
    tags: List[Dict[str, Any]] or Dict[str, Any] = None,
):
    r"""

    A web ACL defines a collection of rules to use to inspect and control web requests. Each rule has an action defined
    (allow, block, or count) for requests that match the statement of the rule. In the web ACL, you assign a default
    action to take (allow, block) for any request that does not match any of the rules. The rules in a web ACL can be
    a combination of the types Rule , RuleGroup , and managed rule group. You can associate a web ACL with one or more
    Amazon Web Services resources to protect. The resources can be an Amazon CloudFront distribution, an Amazon API
    Gateway REST API, an Application Load Balancer, or an AppSync GraphQL API.

    Args:
        name(text): The name of the web ACL. You cannot change the name of a web ACL after you create it.
        scope(text): Specifies whether this is for an Amazon CloudFront distribution or for a regional application.
            A regional application can be an Application Load Balancer (ALB), an Amazon API Gateway REST API,
            or an AppSync GraphQL API.
            To work with CloudFront, you must also specify the Region US East (N. Virginia) as follows:
              * CLI -- Specify the Region when you use the CloudFront scope: --scope=CLOUDFRONT --region=us-east-1 .
              * API and SDKs -- For all calls, use the Region endpoint us-east-1.
        default_action(dict): The action to perform if none of the Rules contained in the WebACL match.
          * Block (dict) -- Specifies that WAF should block requests by default.
          * Allow (dict) -- Specifies that WAF should allow requests by default.
        visibility_config(dict): Defines and enables Amazon CloudWatch metrics and web request sample collection.
          * SampledRequestsEnabled (bool) -- A boolean indicating whether WAF should store a sampling of the web
              requests that match the rules. You can view the sampled requests through the WAF console.
          * CloudWatchMetricsEnabled (bool) -- A boolean indicating whether the associated resource sends metrics to
                Amazon CloudWatch. For the list of available metrics, see WAF Metrics.
          * MetricName (text) -- A name of the Amazon CloudWatch metric. The name can contain only the characters:
                A-Z, a-z, 0-9, - (hyphen), and _ (underscore). The name can be from one to 128 characters long.
                It can't contain whitespace or metric names reserved for WAF, for example "All" and "Default_Action."
        resource_id(text, optional): AWS WAF ID.
        description(text, optional): A description of the web ACL that helps with identification.
        rules(list, optional): The Rule statements used to identify the web requests that you want to allow, block, or
            count. Each rule includes one top-level statement that WAF uses to identify matching web requests, and
            parameters that govern how WAF handles them.
        custom_response_bodies(dict, optional): A map of custom response keys and content bodies. When you create a rule with a
            block action, you can send a custom response to the web request. You define these for the web ACL, and then
            use them in the rules and default actions that you define in the web ACL.
        captcha_config(dict, optional): Specifies how WAF should handle CAPTCHA evaluations for rules that don't have their own
            CaptchaConfig settings. If you don't specify this, WAF uses its default settings for CaptchaConfig.
              * ImmunityTimeProperty (dict) -- Determines how long a CAPTCHA token remains valid after the client
                    successfully solves a CAPTCHA puzzle.
                  * ImmunityTime(integer) -- The amount of time, in seconds, that a CAPTCHA token is valid. The default
                        setting is 300.
        tags(list, optional): A tag associated with an Amazon Web Services resource. Tags are key:value pairs that you can use to
            categorize and manage your resources, for purposes like billing or other management. Typically, the tag key
            represents a category, such as "environment", and the tag value represents a specific value within that
            category, such as "test," "development," or "production". Or you might set the tag key to "customer" and
            the value to the customer name or ID. You can specify one or more tags to add to each Amazon Web Services
            resource, up to 50 tags for a resource.
            list of tags in the format of [{"Key": tag-key, "Value": tag-value}] or dict in the format of
                {tag-key: tag-value}

    Request Syntax:
        [web-acl-name]::
            aws.wafv2.web_acl.present:
            - name: 'string'
            - scope: 'string'
            - default_action: 'dict'
            - description: 'string'
            - rules: ['string']
            - visibility_config: 'dict'
            - custom_response_bodies: 'dict'
            - captcha_config: 'dict'
            - tags:
              - Key: 'string'
                Value: 'string'

    Examples:

        .. code-block:: sls
        demo_waf:
            aws.wafv2.web_acl.present:
            - name: demo_waf.
            - scope: REGIONAL
            - default_action:
                Allow: {}
            - description: waf to support REGIONAL resources.
            - rules:
                - Name: AWS-AWSManagedRulesBotControlRuleSet
                  OverrideAction:
                  None: {}
                  Priority: 1
                  Statement:
                    ManagedRuleGroupStatement:
                    Name: AWSManagedRulesBotControlRuleSet
                    VendorName: AWS
                  VisibilityConfig:
                    CloudWatchMetricsEnabled: false
                    MetricName: AWS-AWSManagedRulesBotControlRuleSet
                    SampledRequestsEnabled: true
            - visibility_config:
                CloudWatchMetricsEnabled: false
                MetricName: test_metric_waf
                SampledRequestsEnabled: true
            - tags:
              - Key: type
                Value: REGIONAL

    Returns:
        Dict[str, Any]

    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    before = None
    lock_token: str = None
    resource_updated: bool = False
    tags_list = None
    tags_dict = None
    if tags is not None:
        if isinstance(tags, Dict):
            tags_list = hub.tool.aws.tag_utils.convert_tag_dict_to_list(tags)
            tags_dict = tags
        else:
            tags_list = tags
            tags_dict = hub.tool.aws.tag_utils.convert_tag_list_to_dict(tags)
    resource_parameters = {
        "Name": name,
        "DefaultAction": default_action,
        "VisibilityConfig": visibility_config,
        "Description": description,
        "Rules": rules,
        "CustomResponseBodies": custom_response_bodies,
        "CaptchaConfig": captcha_config,
        "Tags": tags_list,
    }

    if resource_id:
        ret = await hub.exec.boto3.client.wafv2.get_web_acl(
            ctx, Name=name, Scope=scope, Id=resource_id
        )
        if ret["result"]:
            before = ret["ret"]["WebACL"]
            lock_token = ret["ret"]["LockToken"]

    if before:
        convert_ret = (
            await hub.tool.aws.wafv2.conversion_utils.convert_raw_web_acl_to_present(
                ctx, raw_resource=before, idem_resource_name=name, scope=scope
            )
        )
        result["result"] = convert_ret["result"]
        if not result["result"]:
            result["comment"] = result["comment"] + convert_ret["comment"]
        result["old_state"] = convert_ret["ret"]
        plan_state = copy.deepcopy(result["old_state"])

        # Update web acl
        update_ret = await hub.exec.aws.wafv2.web_acl.update_web_acl(
            ctx,
            name=name,
            raw_resource=before,
            resource_parameters=resource_parameters,
            scope=scope,
            resource_id=resource_id,
            lock_token=lock_token,
        )
        result["comment"] = result["comment"] + update_ret["comment"]
        result["result"] = update_ret["result"]
        resource_updated = resource_updated or bool(update_ret["ret"])

        if update_ret["ret"] and ctx.get("test", False):
            for key in [
                "scope",
                "default_action",
                "visibility_config",
                "description",
                "rules",
                "custom_response_bodies",
                "captcha_config",
            ]:
                if key in update_ret["ret"]:
                    plan_state[key] = update_ret["ret"][key]

        if (tags_list is not None) and (
            not hub.tool.aws.state_comparison_utils.are_lists_identical(
                tags_list,
                hub.tool.aws.tag_utils.convert_tag_dict_to_list(
                    result["old_state"].get("tags")
                ),
            )
        ):
            # Update tags
            update_tag_ret = await hub.exec.aws.wafv2.web_acl.update_web_acl_tags(
                ctx=ctx,
                web_acl_arn=before.get("ARN"),
                old_tags=result["old_state"].get("tags", {}),
                new_tags=tags_dict,
            )
            result["result"] = result["result"] and update_tag_ret["result"]
            result["comment"] = result["comment"] + update_tag_ret["comment"]
            resource_updated = resource_updated or bool(update_tag_ret["ret"])

            if ctx.get("test", False) and update_tag_ret["ret"] is not None:
                plan_state["tags"] = update_tag_ret["ret"]

        if not resource_updated:
            result["comment"] = result["comment"] + (
                f"aws.wafv2.web_acl.present '{name}' has no property to update.",
            )

    else:
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "scope": scope,
                    "default_action": default_action,
                    "visibility_config": visibility_config,
                    "description": description,
                    "rules": rules,
                    "custom_response_bodies": custom_response_bodies,
                    "captcha_config": captcha_config,
                    "tags": tags_dict,
                },
            )
            result["comment"] = (f"Would create aws.wafv2.web_acl {name}",)
            return result

        # Create web acl
        ret = await hub.exec.boto3.client.wafv2.create_web_acl(
            ctx, Scope=scope, **resource_parameters
        )
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = result["comment"] + ret["comment"]
            return result
        result["comment"] = result["comment"] + (f"Created '{name}'",)
        resource_id = ret["ret"]["Summary"]["Id"]

    if ctx.get("test", False):
        result["new_state"] = plan_state

    elif (not before) or resource_updated:
        resource_ret = await hub.exec.boto3.client.wafv2.get_web_acl(
            ctx, Name=name, Scope=scope, Id=resource_id
        )
        if not result["result"]:
            result["comment"] = result["comment"] + ret["comment"]
            return result
        after = resource_ret["ret"]["WebACL"]

        convert_ret = (
            await hub.tool.aws.wafv2.conversion_utils.convert_raw_web_acl_to_present(
                ctx, raw_resource=after, idem_resource_name=name, scope=scope
            )
        )
        result["result"] = convert_ret["result"]
        if not result["result"]:
            result["comment"] = result["comment"] + convert_ret["comment"]
        result["new_state"] = convert_ret["ret"]

    else:
        result["new_state"] = copy.deepcopy(result["old_state"])

    return result


async def absent(
    hub, ctx, name: str, scope: str, resource_id: str = None
) -> Dict[str, Any]:
    r"""
    Deletes the specified WebACL.
    You can only use this if ManagedByFirewallManager is false in the specified WebACL.

    Args:
        name(text): The name of the web ACL. You cannot change the name of a web ACL after you create it.
        resource_id(text): AWS WAF ID. Idem automatically considers this resource being absent if this field is not specified.
        scope(text): Specifies whether this is for an Amazon CloudFront distribution or for a regional application.
            A regional application can be an Application Load Balancer (ALB), an Amazon API Gateway REST API, or an
            AppSync GraphQL API.
            To work with CloudFront, you must also specify the Region US East (N. Virginia) as follows:
            * CLI -- Specify the Region when you use the CloudFront scope: --scope=CLOUDFRONT --region=us-east-1 .
            * API and SDKs -- For all calls, use the Region endpoint us-east-1.

    Request Syntax:
        [web-acl-name]::
            aws.wafv2.web_acl.absent:
            - name: 'string'
            - scope: 'string'

    Examples:

        .. code-block:: sls

            demo_waf::
               aws.wafv2.web_acl.absent:
                - name: demo_waf
                - scope: REGIONAL

    Returns:
        Dict[str, Any]

    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    before = None
    lock_token: str = None

    if resource_id:
        resource = await hub.exec.boto3.client.wafv2.get_web_acl(
            ctx, Name=name, Scope=scope, Id=resource_id
        )
        if resource["result"]:
            before = resource["ret"]["WebACL"]
            lock_token = resource["ret"]["LockToken"]

    if not before:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            "aws.wafv2.web_acl", name
        )

    elif ctx.get("test", False):
        convert_ret = (
            await hub.tool.aws.wafv2.conversion_utils.convert_raw_web_acl_to_present(
                ctx, raw_resource=before, idem_resource_name=name, scope=scope
            )
        )
        result["result"] = convert_ret["result"]
        if not result["result"]:
            result["comment"] = result["comment"] + convert_ret["comment"]
            return result
        result["old_state"] = convert_ret["ret"]
        result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
            "aws.wafv2.web_acl", name
        )
        return result

    else:
        convert_ret = (
            await hub.tool.aws.wafv2.conversion_utils.convert_raw_web_acl_to_present(
                ctx, raw_resource=before, idem_resource_name=name, scope=scope
            )
        )
        result["result"] = convert_ret["result"]
        if not result["result"]:
            result["comment"] = result["comment"] + convert_ret["comment"]
            return result
        result["old_state"] = convert_ret["ret"]

        # Delete web acl
        ret = await hub.exec.boto3.client.wafv2.delete_web_acl(
            ctx, Name=name, Scope=scope, Id=resource_id, LockToken=lock_token
        )
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            return result
        result["comment"] = hub.tool.aws.comment_utils.delete_comment(
            "aws.wafv2.web_acl", name
        )

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""

    Retrieves an WebACLSummary objects for the web ACLs that you manage.

    Examples:

        .. code-block:: bash

            $ idem describe aws.wafv2.web_acl

    Returns:
        Dict[str, Any]

    """
    result = {}
    scope = ["CLOUDFRONT", "REGIONAL"]

    for web_acl_scope in scope:
        ret = await hub.exec.boto3.client.wafv2.list_web_acls(ctx, Scope=web_acl_scope)
        if not ret["result"]:
            hub.log.debug(f"Could not describe web acl {ret['comment']}")
            continue

        for resource in ret["ret"]["WebACLs"]:
            web_acl_name = resource["Name"]
            resource_id = resource["Id"]
            raw_resource = await hub.exec.boto3.client.wafv2.get_web_acl(
                ctx, Name=web_acl_name, Id=resource_id, Scope=web_acl_scope
            )
            if not raw_resource["result"]:
                hub.log.warning(
                    f"Could not get web acl '{web_acl_name}' with error {convert_ret['comment']}"
                )
                continue
            resource_ret = raw_resource["ret"]["WebACL"]
            convert_ret = await hub.tool.aws.wafv2.conversion_utils.convert_raw_web_acl_to_present(
                ctx,
                raw_resource=resource_ret,
                idem_resource_name=web_acl_name,
                scope=web_acl_scope,
            )
            if not convert_ret["result"]:
                hub.log.warning(
                    f"Could not describe web acl '{web_acl_name}' with error {convert_ret['comment']}"
                )
                continue
            translated_resource = convert_ret["ret"]
            result[translated_resource["resource_id"]] = {
                "aws.wafv2.web_acl.present": [
                    {parameter_key: parameter_value}
                    for parameter_key, parameter_value in translated_resource.items()
                ]
            }

    return result



from collections import OrderedDict
from typing import Any
from typing import Dict


def convert_raw_db_subnet_group_to_present(
    hub,
    raw_resource: Dict[str, Any],
    idem_resource_name: str = None,
    tags: Dict[str, Any] = None,
) -> Dict[str, Any]:
    """
    Convert AWS returned data structure to correct idem db_subnet_group present state

     Args:
        hub: required for functions in hub
        raw_resource: The aws response to convert
        idem_resource_name (Text, Optional): The idem name of the resource
        tags (List, Optional): The tags of the resource. Defaults to None.

    Returns: Valid idem state for db_subnet_group of type Dict['string', Any]
    """
    resource_id = raw_resource.get("DBSubnetGroupName")
    raw_resource["Tags"] = tags
    resource_parameters = OrderedDict(
        {
            "DBSubnetGroupDescription": "db_subnet_group_description",
            "Subnets": "subnet_ids",
            "Tags": "tags",
        }
    )
    resource_translated = {"name": idem_resource_name, "resource_id": resource_id}
    for parameter_raw, parameter_present in resource_parameters.items():
        if raw_resource.get(parameter_raw) is not None:
            if parameter_raw == "Subnets":
                resource_translated[parameter_present] = [
                    subnet["SubnetIdentifier"] for subnet in raw_resource.get("Subnets")
                ]
            else:
                resource_translated[parameter_present] = raw_resource.get(parameter_raw)

    return resource_translated



"""
Autogenerated using `pop-create-idem <https://gitlab.com/saltstack/pop/pop-create-idem>`__

hub.exec.boto3.client.iam.create_user
hub.exec.boto3.client.iam.delete_user
hub.exec.boto3.client.iam.get_user
hub.exec.boto3.client.iam.list_users
hub.exec.boto3.client.iam.tag_user
hub.exec.boto3.client.iam.untag_user
hub.exec.boto3.client.iam.update_user
resource = hub.tool.boto3.resource.create(ctx, "iam", "User", name)
hub.tool.boto3.resource.exec(resource, add_group, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, attach_policy, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, create, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, create_access_key_pair, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, create_login_profile, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, create_policy, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, delete, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, detach_policy, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, enable_mfa, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, remove_group, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, update, *args, **kwargs)
"""
import copy
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List

SERVICE = "iam"
RESOURCE = "User"

__contracts__ = ["resource"]
TREQ = {
    "absent": {
        "require": [
            "aws.iam.user_policy_attachment.absent",
        ],
    },
}


async def present(
    hub,
    ctx,
    name: str,
    user_name: str,
    resource_id: str = None,
    path: str = None,
    permissions_boundary: str = None,
    tags: Dict[str, Any]
    or List[
        make_dataclass(
            "Tag",
            [("Key", str, field(default=None)), ("Value", str, field(default=None))],
        )
    ] = None,
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Creates a new IAM user for your Amazon Web Services account.  For information about quotas for the number of IAM
    users you can create, see IAM and STS quotas in the IAM User Guide.

    Args:
        name(Text): An Idem name of the resource.
        resource_id(Text, optional): AWS IAM username
        user_name(Text): The name of the user to create. IAM user, group, role, and policy names must be unique within
            the account. Names are not distinguished by case. For example, you cannot create resources named
            both "MyResource" and "myresource".
        path(Text, optional):  The path for the user name. For more information about paths, see IAM identifiers in the IAM
            User Guide. This parameter is optional. If it is not included, it defaults to a slash (/). This
            parameter allows (through its regex pattern) a string of characters consisting of either a
            forward slash (/) by itself or a string that must begin and end with forward slashes. In
            addition, it can contain any ASCII character from the ! (\u0021) through the DEL character
            (\u007F), including most punctuation characters, digits, and upper and lowercased letters. Defaults to None.
        permissions_boundary(Text, optional): The ARN of the policy that is used to set the permissions boundary for the user. Defaults to None.
        tags(Dict or List, optional): Dict in the format of {tag-key: tag-value} or List of tags in the format of
            [{"Key": tag-key, "Value": tag-value}] to associate with the user.
            Each tag consists of a key name and an  associated value. Defaults to None.
            * (Key): The key name that can be used to look up or retrieve the associated value. For example,
                Department or Cost Center are common choices.
            * (Value): The value associated with this tag. For example, tags with a key name of Department could have
                values such as Human Resources, Accounting, and Support. Tags with a key name of Cost Center
                might have values that consist of the number associated with the different cost centers in your
                company. Typically, many resources have tags with the same key name but with different values.
                Amazon Web Services always interprets the tag Value as a single string. If you need to store an
                array, you can store comma-separated values in the string. However, you must interpret the value
                in your code.

    Request Syntax:
        user-name:
              aws.iam.user.present:
                - user_name: 'string'
                - resource_id: 'string'
                - path: 'string'
                - permissions_boundary: 'string'
                - tags:
                    - Key: 'string'
                      Value: 'string'

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            user1:
              aws.iam.user.present:
                - user_name: iam-user-1
                - resource_id: iam-user-1
                - path: /
                - permissions_boundary: arn:aew12k3r4kr9efg9
                - tags:
                    - Key: test-key
                      Value: test-value
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    before = None
    is_user_updated = False
    if resource_id:
        resource = hub.tool.boto3.resource.create(
            ctx, SERVICE, RESOURCE, name=resource_id
        )
        before = await hub.tool.boto3.resource.describe(resource)

    if isinstance(tags, List):
        tags = hub.tool.aws.tag_utils.convert_tag_list_to_dict(tags)

    if before:
        result[
            "old_state"
        ] = hub.tool.aws.iam.conversion_utils.convert_raw_user_to_present(
            raw_resource=before, idem_resource_name=name
        )
        plan_state = copy.deepcopy(result["old_state"])

        update_ret = await hub.exec.aws.iam.user.update_user(
            ctx, before=result["old_state"], user_name=user_name, path=path
        )
        result["comment"] = result["comment"] + update_ret["comment"]
        result["result"] = update_ret["result"]
        if update_ret["ret"] and "user_name" in update_ret["ret"]:
            resource_id = user_name
        is_user_updated = bool(update_ret["ret"])
        if update_ret["ret"] and ctx.get("test", False):
            result["comment"] = (f"Would update aws.iam.user '{name}'",)
            if "user_name" in update_ret["ret"]:
                plan_state["user_name"] = update_ret["ret"]["user_name"]
                plan_state["resource_id"] = update_ret["ret"]["user_name"]
            if "path" in update_ret["ret"]:
                plan_state["path"] = update_ret["ret"]["path"]

        if tags is not None and tags != before.get("Tags", None):
            # Update tags
            update_ret = await hub.exec.aws.iam.user.update_user_tags(
                ctx,
                user_name=resource_id,
                old_tags=result["old_state"].get("tags", {}),
                new_tags=tags,
            )
            result["result"] = result["result"] and update_ret["result"]
            result["comment"] = result["comment"] + update_ret["comment"]
            is_user_updated = is_user_updated or bool(update_ret["result"])
            if ctx.get("test", False) and update_ret["ret"] is not None:
                plan_state["tags"] = update_ret["ret"]
        if not is_user_updated:
            result["comment"] = result["comment"] + (f"{name} already exists",)
    else:
        try:
            if ctx.get("test", False):
                result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                    enforced_state={},
                    desired_state={
                        "name": name,
                        "user_name": user_name,
                        "arn": "arn_known_after_present",
                        "permissions_boundary": permissions_boundary,
                        "path": path,
                        "tags": tags,
                    },
                )
                result["comment"] = (f"Would create aws.iam.user '{name}'",)
                return result
            # Attempt to create the User
            ret = await hub.exec.boto3.client.iam.create_user(
                ctx,
                Path=path,
                UserName=user_name,
                PermissionsBoundary=permissions_boundary,
                Tags=hub.tool.aws.tag_utils.convert_tag_dict_to_list(tags)
                if tags
                else None,
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                return result
            result["comment"] = (f"Created aws.iam.user '{name}'",)
            resource_id = user_name
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = result["comment"] + (f"{e.__class__.__name__}: {e}",)
            result["result"] = False

    try:
        if ctx.get("test", False):
            result["new_state"] = plan_state
        elif (not before) or is_user_updated:
            resource = hub.tool.boto3.resource.create(
                ctx, SERVICE, RESOURCE, name=resource_id
            )
            after = await hub.tool.boto3.resource.describe(resource)
            result[
                "new_state"
            ] = hub.tool.aws.iam.conversion_utils.convert_raw_user_to_present(
                raw_resource=after, idem_resource_name=name
            )
        else:
            result["new_state"] = copy.deepcopy(result["old_state"])
    except Exception as e:
        result["comment"] = result["comment"] + (str(e),)
        result["result"] = False

    return result


async def absent(hub, ctx, name: str, resource_id: str) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Deletes the specified IAM user. Unlike the Amazon Web Services Management Console, when you delete a user
    programmatically, you must delete the items attached to the user manually, or the deletion fails. For more
    information, see Deleting an IAM user. Before attempting to delete a user, remove the following items:
    Password (DeleteLoginProfile)   Access keys (DeleteAccessKey)   Signing certificate (DeleteSigningCertificate)
    SSH public key (DeleteSSHPublicKey)   Git credentials (DeleteServiceSpecificCredential)   Multi-factor
    authentication (MFA) device (DeactivateMFADevice, DeleteVirtualMFADevice)   Inline policies (DeleteUserPolicy)
    Attached managed policies (DetachUserPolicy)   Group memberships (RemoveUserFromGroup)

    Args:
        name(Text): A name, ID to identify the resource.
        resource_id(Text, optional): The name of the user to delete. This parameter allows (through its regex pattern) a string of
            characters consisting of upper and lowercase alphanumeric characters with no spaces. You can
            also include any of the following characters: _+=,.@-.

    Request syntax:

        user-1:
              aws.iam.user.absent:
                - name: 'string'
                - resource_id: 'string'

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            user-1:
              aws.iam.user.absent:
                - name: user-1
                - resource_id: user-1
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    resource = hub.tool.boto3.resource.create(ctx, SERVICE, RESOURCE, resource_id)
    before = await hub.tool.boto3.resource.describe(resource)

    if not before:
        result["comment"] = (f"'{name}' already absent",)
    elif ctx.get("test", False):
        result[
            "old_state"
        ] = hub.tool.aws.iam.conversion_utils.convert_raw_user_to_present(
            raw_resource=before, idem_resource_name=name
        )
        result["comment"] = (f"Would delete aws.iam.user '{name}'",)
        return result
    else:
        result[
            "old_state"
        ] = hub.tool.aws.iam.conversion_utils.convert_raw_user_to_present(
            raw_resource=before, idem_resource_name=name
        )
        try:
            # Attempt to delete user ssh key
            user_ssh_public_keys_list = (
                await hub.exec.boto3.client.iam.list_ssh_public_keys(ctx, UserName=name)
            )
            if user_ssh_public_keys_list["result"]:
                attached_user_ssh_public_keys_list = user_ssh_public_keys_list[
                    "ret"
                ].get("SSHPublicKeys")
                for ssh_key in attached_user_ssh_public_keys_list:
                    ret_delete = await hub.exec.boto3.client.iam.delete_ssh_public_key(
                        ctx, UserName=name, SSHPublicKeyId=ssh_key.get("SSHPublicKeyId")
                    )
                    if not ret_delete["result"]:
                        result["result"] = False
                        result["comment"] = ret_delete["comment"]
                        hub.log.debug(
                            f"Could not delete user ssh key {ret_delete['comment']}"
                        )
                        return result
            else:
                if (
                    user_ssh_public_keys_list["comment"]
                    and "NoSuchEntityException"
                    in user_ssh_public_keys_list["comment"][0]
                ):
                    result["comment"] = (
                        f"No user_ssh_key associated with aws.iam.user '{name}'",
                    )
                else:
                    result["result"] = False
                    result["comment"] = user_ssh_public_keys_list["comment"]
                    return result

            # Attempt to delete the User
            ret = await hub.exec.boto3.client.iam.delete_user(
                ctx,
                UserName=resource_id,
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                return result
            result["comment"] = f"Deleted aws.iam.user '{name}'"
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = result["comment"] + (f"{e.__class__.__name__}: {e}",)
            result["result"] = False
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""
    **Autogenerated function**

    Describe the resource in a way that can be recreated/managed with the corresponding "present" function


    Lists the IAM users that have the specified path prefix. If no path prefix is specified, the operation returns
    all users in the Amazon Web Services account. If there are none, the operation returns an empty list.  IAM
    resource-listing operations return a subset of the available attributes for the resource. For example, this
    operation does not return tags, even though they are an attribute of the returned object. To view all of the
    information for a user, see GetUser.  You can paginate the results using the MaxItems and Marker parameters.


    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: bash

            $ idem describe aws.iam.user
    """
    result = {}
    ret = await hub.exec.boto3.client.iam.list_users(ctx)

    if not ret["result"]:
        hub.log.debug(f"Could not describe user {ret['comment']}")
        return {}

    for user in ret["ret"]["Users"]:
        resource_name = user.get("UserName")
        # This is required to get tags for each user
        boto2_resource = hub.tool.boto3.resource.create(
            ctx, "iam", "User", resource_name
        )
        resource = await hub.tool.boto3.resource.describe(boto2_resource)
        resource_id = resource.get("UserName")
        translated_resource = (
            hub.tool.aws.iam.conversion_utils.convert_raw_user_to_present(
                raw_resource=resource, idem_resource_name=resource_id
            )
        )
        resource_key = f"iam-user-{resource_id}"
        result[resource_key] = {
            "aws.iam.user.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in translated_resource.items()
            ]
        }

    return result



import copy
from collections import OrderedDict
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    db_instance_class: str,
    engine: str,
    resource_id: str = None,
    db_name: str = None,
    allocated_storage: int = None,
    master_username: str = None,
    master_user_password: str = None,
    db_security_groups: List = None,
    vpc_security_group_ids: List = None,
    availability_zone: str = None,
    db_subnet_group_name: str = None,
    preferred_maintenance_window: str = None,
    db_parameter_group_name: str = None,
    backup_retention_period: int = None,
    preferred_backup_window: str = None,
    port: int = None,
    multi_az: bool = None,
    engine_version: str = None,
    auto_minor_version_upgrade: bool = None,
    license_model: str = None,
    iops: int = None,
    option_group_name: str = None,
    character_set_name: str = None,
    nchar_character_set_name: str = None,
    publicly_accessible: bool = None,
    tags: List[Dict[str, Any]] or Dict[str, Any] = None,
    db_cluster_identifier: str = None,
    storage_type: str = None,
    tde_credential_arn: str = None,
    tde_credential_password: str = None,
    storage_encrypted: bool = None,
    kms_key_id: str = None,
    domain: str = None,
    copy_tags_to_snapshot: bool = None,
    monitoring_interval: int = None,
    monitoring_role_arn: str = None,
    domain_iam_role_name: str = None,
    promotion_tier: int = None,
    timezone: str = None,
    enable_iam_database_authentication: bool = None,
    enable_performance_insights: bool = None,
    performance_insights_kms_key_id: str = None,
    performance_insights_retention_period: int = None,
    enable_cloudwatch_logs_exports: List = None,
    processor_features: List = None,
    deletion_protection: bool = None,
    max_allocated_storage: int = None,
    enable_customer_owned_ip: bool = None,
    custom_iam_instance_profile: str = None,
    backup_target: str = None,
    apply_immediately: bool = None,
    allow_major_version_upgrades: bool = None,
    ca_certificate_identifier: str = None,
    cloudwatch_logs_export_configuration: Dict = None,
    use_default_processor_features: bool = None,
    certificate_rotation_restart: bool = None,
    replica_mode: str = None,
    aws_backup_recovery_point_arn: str = None,
    automation_mode: str = None,
    resume_full_automation_mode_minutes: int = None,
    timeout: Dict = None,
) -> Dict[str, Any]:
    r'''
    **Autogenerated function**

    Creates a new DB instance.

    Args:
        hub:
        ctx:
        name(Text): Idem name to identify the resource.
        allocated_storage(int): The amount of storage in gibibytes (GiB) to allocate for the DB instance
        db_instance_class(Text): The compute and memory capacity of the DB instance, for example db.m4.large
        engine(Text): The name of the database engine to be used for this instance.
        resource_id(Text, optional): AWS Id of the resource
        db_name(Text, optional): Name of the database
        master_username(Text, optional): The name for the master user.Not required for aurora
        master_user_password(Text, optional): The password for the master user. The password can include any printable
                                             ASCII character except "/", """, or "@".Not required for aurora
        db_security_groups(list, optional): A list of DB security groups to associate with this DB instance.
        vpc_security_group_ids(list, optional): A list of DB security groups to associate with this DB instance.
        availability_zone(Text, optional): The Availability Zone (AZ) where the database will be created.
        db_subnet_group_name(Text, optional): A DB subnet group to associate with this DB instance.
        preferred_maintenance_window(Text, optional): The time range each week during which system maintenance can occur,
                                                     in Universal Coordinated Time (UTC).
        db_parameter_group_name(Text, optional): The name of the DB parameter group to associate with this DB instance.
                                                If you do not specify a value,then the default DB parameter group for
                                                the specified DB engine and version is used.
        backup_retention_period(int, optional): The number of days for which automated backups are retained. Setting
                                                this parameter to a positive number enables backups. Setting this
                                                parameter to 0 disables automated backups.
        preferred_backup_window(Text, optional): The daily time range during which automated backups are created if
                                                automated backups are enabled,using the BackupRetentionPeriod parameter.
        port(int, optional): The port number on which the database accepts connections.
        multi_az(bool, optional): A value that indicates whether the DB instance is a Multi-AZ deployment. You can't set
                                  the AvailabilityZone parameter if the DB instance is a Multi-AZ deployment.
        engine_version(Text, optional): The version number of the database engine to use.
        auto_minor_version_upgrade(bool, optional): A value that indicates whether minor engine upgrades are applied
                                                    automatically to the DB instance during the maintenance window. By
                                                    default, minor engine upgrades are applied automatically.
        license_model(Text, optional): License model information for this DB instance.
        iops(int, optional): The amount of Provisioned IOPS (input/output operations per second) to be initially
                             allocated for the DB instance.
        option_group_name(Text, optional): A value that indicates that the DB instance should be associated with the
                                          specified option group.
        character_set_name(Text, optional): For supported engines, this value indicates that the DB instance should be
                                           associated with the specified CharacterSet .
        nchar_character_set_name(Text, optional): The name of the NCHAR character set for the Oracle DB instance.
        publicly_accessible(bool, optional): A value that indicates whether the DB instance is publicly accessible.
        tags(List or Dict, optional): Tags to assign to the DB instance. List of tags in the format of [{"Key": tag-key, "Value": tag-value}] or dict in the format of
                                      {tag-key: tag-value}
        db_cluster_identifier(Text, optional): The identifier of the DB cluster that the instance will belong to.
        storage_type(Text, optional): Specifies the storage type to be associated with the DB instance.
        tde_credential_arn(Text, optional): The ARN from the key store with which to associate the instance for TDE
                                           encryption.
        tde_credential_password(Text, optional): The password for the given ARN from the key store in order to access the
                                                device.
        storage_encrypted(bool, optional): A value that indicates whether the DB instance is encrypted. By default, it
                                           isn't encrypted
        kms_key_id(Text, optional): The Amazon Web Services KMS key identifier for an encrypted DB instance.
        domain(Text, optional): The Active Directory, directory ID to create the DB instance in
        copy_tags_to_snapshot(Text, optional): A value that indicates whether to copy tags from the DB instance to
                                              snapshots of the DB instance. By default, tags are not copied.
        monitoring_interval(int, optional): The interval, in seconds, between points when Enhanced Monitoring metrics
                                            are collected for the DB instance.To disable collection of Enhanced
                                            Monitoring metrics, specify 0. The default is 0.
                                            If MonitoringRoleArn is specified, then you must set MonitoringInterval to
                                            a value other than 0.
        monitoring_role_arn(Text, optional): The ARN for the IAM role that permits RDS to send enhanced monitoring metrics
                                            to Amazon CloudWatch Logs.
        domain_iam_role_name(Text, optional): Specify the name of the IAM role to be used when making API calls to the
                                             Directory Service.
        promotion_tier(int, optional): A value that specifies the order in which an Aurora Replica is promoted to the
                                       primary instance after a failure of the existing primary instance.
        timezone(Text, optional): The time zone of the DB instance.
        enable_iam_database_authentication(bool, optional): A value that indicates whether to enable mapping of Amazon
                                                            Web Services Identity and Access Management (IAM) accounts
                                                            to database accounts. By default, mapping isn't enabled.
        enable_performance_insights(bool, optional): A value that indicates whether to enable Performance Insights for the DB instance.
        performance_insights_kms_key_id(Text, optional): The Amazon Web Services KMS key identifier for encryption of
                                                        Performance Insights data.
        performance_insights_retention_period(int, optional): The amount of time, in days, to retain Performance
                                                              Insights data.
        enable_cloudwatch_logs_exports(List, optional):The list of log types that need to be enabled for exporting to
                                                       CloudWatch Logs. The values in the list depend on the DB engine.
        processor_features(List, optional): The number of CPU cores and the number of threads per core for the DB
                                            instance class of the DB instance.
        deletion_protection(bool, optional):A value that indicates whether the DB instance has deletion protection
                                            enabled. The database can't be deleted when deletion protection is enabled.
                                            By default, deletion protection isn't enabled.
        max_allocated_storage(int, optional): The upper limit in gigabytes (GiB) to which Amazon RDS can automatically
                                              scale the storage of the DB instance.
        enable_customer_owned_ip(bool, optional): A value that indicates whether to enable a customer-owned IP address
                                                  (CoIP) for an RDS on Outposts DB instance.
        custom_iam_instance_profile(Text, optional): The instance profile associated with the underlying Amazon EC2
                                                    instance of an RDS Custom DB instance.
        backup_target(Text, optional):  Value to specify where automated backups and manual snapshots are stored.

        # parameters for the modify function
        apply_immediately(bool, optional): A value that indicates whether the modifications in this request and any
                                           pending modifications are asynchronously applied as soon as possible,
                                           regardless of the PreferredMaintenanceWindow setting for the DB instance.
                                           By default, this parameter is disabled.
        allow_major_version_upgrades(bool, optional): A value that indicates whether major version upgrades are allowed.
                                                      Changing this parameter doesn't result in an outage and the change
                                                      is asynchronously applied as soon as possible.
        cloudwatch_logs_export_configuration(Dict, optional): The configuration setting for the log types to be enabled
                                                              for export to CloudWatch Logs for a specific DB instance.
        ca_certificate_identifier(Text, optional): The identifier of the CA certificate for the DB instance.
        use_default_processor_features(bool, optional): A value that indicates whether the DB instance class of the DB
                                                        instance uses its default processor features.
        certificate_rotation_restart(bool, optional): A value that indicates whether the DB instance is restarted when
                                                      you rotate your SSL/TLS certificate.
        replica_mode(Text, optional):A value that sets the open mode of a replica database to either mounted or read-only
        aws_backup_recovery_point_arn(Text, optional): The Amazon Resource Name (ARN) of the recovery point in Amazon Web
                                                      Services Backup.
        automation_mode(Text, optional): The automation mode of the RDS Custom DB instance: full or all paused .
        resume_full_automation_mode_minutes(int, optional):The number of minutes to pause the automation. When the time
                                                            period ends, RDS Custom resumes full automation
        timeout(Dict, optional): Timeout configuration for create/update of AWS DB Cluster.
            * create (Dict) -- Timeout configuration for creating DB Instance
                * delay(int) -- The amount of time in seconds to wait between attempts.Defaults to 30
                * max_attempts(int) -- Customized timeout configuration containing delay and max attempts.Defaults to 60
            * update (Dict) -- Timeout configuration for updating DB Instance
                * delay(int) -- The amount of time in seconds to wait between attempts.Defaults to 30
                * max_attempts(int) -- Customized timeout configuration containing delay and max attempts.Defaults to 60


    Request Syntax:
        [db-instance-name]:
          aws.rds.db_instance.present:
          - db_instance_class: 'string'
          - engine: 'string'
          - resource_id: 'string'
          - availability_zone: 'string'
          - preferred_maintenance_window: 'string'
          - multi_az: 'bool'
          - engine_version: 'string'
          - auto_minor_version_upgrade: 'bool'
          - license_model: 'string'
          - publicly_accessible: 'bool'
          - db_cluster_identifier: 'string'
          - storage_type: 'string'
          - storage_encrypted: 'bool'
          - copy_tags_to_snapshot: 'bool'
          - monitoring_interval: 'int'
          - monitoring_role_arn: 'string'
          - promotion_tier: 'int'
          - enable_performance_insights: 'bool'
          - performance_insights_kms_key_id: 'string'
          - performance_insights_retention_period: 'int'
          - backup_target: 'string'
          - vpc_security_group_id: 'List'
          - db_subnet_group_name: 'string'
          - db_parameter_group_name: 'string'
          - option_group_name: 'string'
          - tags: 'List'
          - timeout: 'Dict'

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

              instance-1:
                aws.rds.db_instance.present:
                - db_instance_class: db.r5.large
                - engine: aurora-postgresql
                - availability_zone: us-east-2b
                - preferred_maintenance_window: sat:04:15-sat:04:45
                - multi_az: false
                - engine_version: '12.7'
                - auto_minor_version_upgrade: true
                - license_model: postgresql-license
                - publicly_accessible: false
                - db_cluster_identifier: db-cluster-1
                - storage_type: aurora
                - storage_encrypted: true
                - copy_tags_to_snapshot: false
                - monitoring_interval: 60
                - monitoring_role_arn: arn:aws:iam::537227425989:role/rds-monitoring-role
                - promotion_tier: 1
                - enable_performance_insights: true
                - performance_insights_kms_key_id: arn:aws:kms:us-east-2:537227425989:key/e9e79921-8dda-48d7-afd7-38a64dd8e9b1
                - performance_insights_retention_period: 7
                - backup_target: region
                - vpc_security_group_id:
                  - sg-f5eeba9c
                - db_subnet_group_name: default
                - db_parameter_group_name: default.aurora-postgresql12
                - option_group_name: default:aurora-postgresql-12
                - tags:
                  - Key: name
                    Value: value
                - timeout:
                    create:
                      delay: 10
                      max_attempts: 30

    '''

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    params_to_modify = {}
    plan_state = None
    before = None
    resource_updated = False

    if resource_id:
        before = await hub.exec.boto3.client.rds.describe_db_instances(
            ctx, DBInstanceIdentifier=resource_id
        )
    tags = (
        hub.tool.aws.tag_utils.convert_tag_list_to_dict(tags)
        if isinstance(tags, List)
        else tags
    )
    if before and before.get("ret"):
        db_instance_arn = before["ret"]["DBInstances"][0].get("DBInstanceArn")
        ret_tags = await hub.exec.boto3.client.rds.list_tags_for_resource(
            ctx, ResourceName=db_instance_arn
        )
        if not ret_tags["result"]:
            result["comment"] = ret_tags["comment"]
            result["result"] = False
            return result
        result[
            "old_state"
        ] = await hub.tool.aws.rds.conversion_utils.convert_raw_db_instance_to_present(
            raw_resource=before["ret"]["DBInstances"][0],
            raw_resource_tags=ret_tags,
            idem_resource_name=name,
        )
        plan_state = copy.deepcopy(result["old_state"])
        old_tags = plan_state.get("tags")

        if tags is not None:
            update_tags_ret = await hub.exec.aws.rds.tag.update_rds_tags(
                ctx,
                resource_arn=db_instance_arn,
                old_tags=old_tags,
                new_tags=tags,
            )
            result["comment"] = update_tags_ret["comment"]
            if not update_tags_ret["result"]:
                result["result"] = False
                return result
            resource_updated = update_tags_ret["result"]
            if ctx.get("test", False) and update_tags_ret["result"]:
                plan_state["tags"] = update_tags_ret["ret"]

        modify_params = OrderedDict(
            {
                "AllocatedStorage": "allocated_storage",
                "DBInstanceClass": "db_instance_class",
                "MasterUserPassword": "master_user_password",
                "DBSecurityGroups": "db_security_groups",
                "VpcSecurityGroupIds": "vpc_security_group_ids",
                "ApplyImmediately": "apply_immediately",
                "DBSubnetGroupName": "db_subnet_group_name",
                "PreferredMaintenanceWindow": "preferred_maintenance_window",
                "DBParameterGroupName": "db_parameter_group_name",
                "BackupRetentionPeriod": "backup_retention_period",
                "PreferredBackupWindow": "preferred_backup_window",
                "MultiAZ": "multi_az",
                "EngineVersion": "engine_version",
                "AllowMajorVersionUpgrade": "allow_major_version_upgrades",
                "AutoMinorVersionUpgrade": "auto_minor_version_upgrade",
                "LicenseModel": "license_model",
                "Iops": "iops",
                "OptionGroupName": "option_group_name",
                "StorageType": "storage_type",
                "TdeCredentialArn": "tde_credential_arn",
                "TdeCredentialPassword": "tde_credential_password",
                "CACertificateIdentifier": "ca_certificate_identifier",
                "Domain": "domain",
                "CopyTagsToSnapshot": "copy_tags_to_snapshot",
                "MonitoringInterval": "monitoring_interval",
                "DBPortNumber": "port",
                "PubliclyAccessible": "publicly_accessible",
                "MonitoringRoleArn": "monitoring_role_arn",
                "DomainIAMRoleName": "domain_iam_role_name",
                "PromotionTier": "promotion_tier",
                "EnableIAMDatabaseAuthentication": "enable_iam_database_authentication",
                "EnablePerformanceInsights": "enable_performance_insights",
                "PerformanceInsightsKMSKeyId": "performance_insights_kms_key_id",
                "PerformanceInsightsRetentionPeriod": "performance_insights_retention_period",
                "CloudwatchLogsExportConfiguration": "cloudwatch_logs_export_configuration",
                "ProcessorFeatures": "processor_features",
                "UseDefaultProcessorFeatures": "use_default_processor_features",
                "DeletionProtection": "deletion_protection",
                "MaxAllocatedStorage": "max_allocated_storage",
                "CertificateRotationRestart": "certificate_rotation_restart",
                "ReplicaMode": "replica_mode",
                "EnableCustomerOwnedIp": "enable_customer_owned_ip",
                "AwsBackupRecoveryPointArn": "aws_backup_recovery_point_arn",
                "AutomationMode": "automation_mode",
                "ResumeFullAutomationModeMinutes": "resume_full_automation_mode_minutes",
            }
        )
        for parameter_raw, parameter_present in modify_params.items():
            # Add to modify list only if parameter is changed
            if (
                locals()[parameter_present] is not None
                and result["old_state"].get(parameter_present)
                != locals()[parameter_present]
            ):
                params_to_modify[parameter_raw] = locals()[parameter_present]

        if params_to_modify:
            if ctx.get("test", False):
                result["comment"] = result["comment"] + (
                    f"Would update aws.rds.db_instance '{name}'",
                )
                for key, value in params_to_modify.items():
                    plan_state[modify_params.get(key)] = value
            else:
                modify_ret = await hub.exec.boto3.client.rds.modify_db_instance(
                    ctx, DBInstanceIdentifier=resource_id, **params_to_modify
                )
                if not modify_ret["result"]:
                    result["comment"] = result["comment"] + modify_ret["comment"]
                    result["result"] = False
                    return result
                # Waiting for the DBInstances status to be available after updates
                waiter_config = hub.tool.aws.waiter_utils.create_waiter_config(
                    default_delay=30,
                    default_max_attempts=60,
                    timeout_config=timeout.get("update") if timeout else None,
                )
                try:
                    await hub.tool.boto3.client.wait(
                        ctx,
                        "rds",
                        "db_instance_available",
                        None,
                        30,
                        DBInstanceIdentifier=resource_id,
                        WaiterConfig=waiter_config,
                    )
                except Exception as e:
                    result["comment"] = result["comment"] + (str(e),)
                    result["result"] = False

                resource_updated = resource_updated or modify_ret["result"]
                result["comment"] = result["comment"] + (
                    f"Updated aws.rds.db_instance '{name}'.updated parameters '{params_to_modify}'",
                )

    else:
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "db_name": db_name,
                    "allocated_storage": allocated_storage,
                    "db_instance_class": db_instance_class,
                    "engine": engine,
                    "master_username": master_username,
                    "master_user_password": master_user_password,
                    "db_security_groups": db_security_groups,
                    "vpc_security_group_ids": vpc_security_group_ids,
                    "availability_zone": availability_zone,
                    "db_subnet_group_name": db_subnet_group_name,
                    "preferred_maintenance_window": preferred_maintenance_window,
                    "db_parameter_group_name": db_parameter_group_name,
                    "backup_retention_period": backup_retention_period,
                    "preferred_backup_window": preferred_backup_window,
                    "port": port,
                    "multi_az": multi_az,
                    "engine_version": engine_version,
                    "auto_minor_version_upgrade": auto_minor_version_upgrade,
                    "license_model": license_model,
                    "iops": iops,
                    "option_group_name": option_group_name,
                    "character_set_name": character_set_name,
                    "nchar_character_set_name": nchar_character_set_name,
                    "publicly_accessible": publicly_accessible,
                    "tags": tags,
                    "db_cluster_identifier": db_cluster_identifier,
                    "storage_type": storage_type,
                    "tde_credential_arn": tde_credential_arn,
                    "tde_credential_password": tde_credential_password,
                    "storage_encrypted": storage_encrypted,
                    "kms_key_id": kms_key_id,
                    "domain": domain,
                    "copy_tags_to_snapshot": copy_tags_to_snapshot,
                    "monitoring_interval": monitoring_interval,
                    "monitoring_role_arn": monitoring_role_arn,
                    "domain_iam_role_name": domain_iam_role_name,
                    "promotion_tier": promotion_tier,
                    "timezone": timezone,
                    "enable_iam_database_authentication": enable_iam_database_authentication,
                    "enable_performance_insights": enable_performance_insights,
                    "performance_insights_kms_key_id": performance_insights_kms_key_id,
                    "performance_insights_retention_period": performance_insights_retention_period,
                    "enable_cloudwatch_logs_exports": enable_cloudwatch_logs_exports,
                    "processor_features": processor_features,
                    "deletion_protection": deletion_protection,
                    "max_allocated_storage": max_allocated_storage,
                    "enable_customer_owned_ip": enable_customer_owned_ip,
                    "custom_iam_instance_profile": custom_iam_instance_profile,
                    "backup_target": backup_target,
                },
            )
            result["comment"] = (f"Would create aws.rds.db_instance '{name}'",)
            return result
        ret = await hub.exec.boto3.client.rds.create_db_instance(
            ctx,
            DBInstanceIdentifier=name,
            DBName=db_name,
            AllocatedStorage=allocated_storage,
            DBInstanceClass=db_instance_class,
            Engine=engine,
            MasterUsername=master_username,
            MasterUserPassword=master_user_password,
            DBSecurityGroups=db_security_groups,
            VpcSecurityGroupIds=vpc_security_group_ids,
            AvailabilityZone=availability_zone,
            DBSubnetGroupName=db_subnet_group_name,
            PreferredMaintenanceWindow=preferred_maintenance_window,
            DBParameterGroupName=db_parameter_group_name,
            BackupRetentionPeriod=backup_retention_period,
            PreferredBackupWindow=preferred_backup_window,
            Port=port,
            MultiAZ=multi_az,
            EngineVersion=engine_version,
            AutoMinorVersionUpgrade=auto_minor_version_upgrade,
            LicenseModel=license_model,
            Iops=iops,
            OptionGroupName=option_group_name,
            CharacterSetName=character_set_name,
            NcharCharacterSetName=nchar_character_set_name,
            PubliclyAccessible=publicly_accessible,
            Tags=hub.tool.aws.tag_utils.convert_tag_dict_to_list(tags)
            if tags
            else None,
            DBClusterIdentifier=db_cluster_identifier,
            StorageType=storage_type,
            TdeCredentialArn=tde_credential_arn,
            TdeCredentialPassword=tde_credential_password,
            StorageEncrypted=storage_encrypted,
            KmsKeyId=kms_key_id,
            Domain=domain,
            CopyTagsToSnapshot=copy_tags_to_snapshot,
            MonitoringInterval=monitoring_interval,
            MonitoringRoleArn=monitoring_role_arn,
            DomainIAMRoleName=domain_iam_role_name,
            PromotionTier=promotion_tier,
            Timezone=timezone,
            EnableIAMDatabaseAuthentication=enable_iam_database_authentication,
            EnablePerformanceInsights=enable_performance_insights,
            PerformanceInsightsKMSKeyId=performance_insights_kms_key_id,
            PerformanceInsightsRetentionPeriod=performance_insights_retention_period,
            EnableCloudwatchLogsExports=enable_cloudwatch_logs_exports,
            ProcessorFeatures=processor_features,
            DeletionProtection=deletion_protection,
            MaxAllocatedStorage=max_allocated_storage,
            EnableCustomerOwnedIp=enable_customer_owned_ip,
            CustomIamInstanceProfile=custom_iam_instance_profile,
            BackupTarget=backup_target,
        )
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            return result
        resource_id = ret["ret"]["DBInstance"]["DBInstanceIdentifier"]
        # Waiting until the DBInstance creation
        waiter_config = hub.tool.aws.waiter_utils.create_waiter_config(
            default_delay=30,
            default_max_attempts=60,
            timeout_config=timeout.get("create") if timeout else None,
        )
        try:
            await hub.tool.boto3.client.wait(
                ctx,
                "rds",
                "db_instance_available",
                DBInstanceIdentifier=resource_id,
                WaiterConfig=waiter_config,
            )
        except Exception as e:
            result["comment"] = result["comment"] + (str(e),)
            result["result"] = False
        result["comment"] = result["comment"] + (
            f"Created aws.rds.db_instance '{name}'",
        )

    if ctx.get("test", False):
        result["new_state"] = plan_state
    elif (not before) or resource_updated:
        after = await hub.exec.boto3.client.rds.describe_db_instances(
            ctx, DBInstanceIdentifier=resource_id
        )
        db_instance_arn = after["ret"]["DBInstances"][0].get("DBInstanceArn")
        ret_tags = await hub.exec.boto3.client.rds.list_tags_for_resource(
            ctx, ResourceName=db_instance_arn
        )
        if not ret_tags["result"]:
            result["comment"] = result["comment"] + ret_tags["comment"]
            result["result"] = False
        result[
            "new_state"
        ] = await hub.tool.aws.rds.conversion_utils.convert_raw_db_instance_to_present(
            raw_resource=after["ret"]["DBInstances"][0],
            raw_resource_tags=ret_tags,
            idem_resource_name=name,
        )
    else:
        result["new_state"] = copy.deepcopy(result["old_state"])

    return result


async def absent(
    hub,
    ctx,
    name: str,
    resource_id: str,
    skip_final_snapshot: bool = None,
    final_db_snapshot_identifier: str = None,
    timeout: Dict = None,
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    The DeleteDBInstance action deletes a previously provisioned DB instance. When you delete a DB instance, all
    automated backups for that instance are deleted and can't be recovered. Manual DB snapshots of the DB instance
    to be deleted by DeleteDBInstance are not deleted.  If you request a final DB snapshot the status of the Amazon
    RDS DB instance is deleting until the DB snapshot is created. The API action DescribeDBInstance is used to
    monitor the status of this operation. The action can't be canceled or reverted once submitted.  When a DB
    instance is in a failure state and has a status of failed, incompatible-restore, or incompatible-network, you
    can only delete it when you skip creation of the final snapshot with the SkipFinalSnapshot parameter. If the
    specified DB instance is part of an Amazon Aurora DB cluster, you can't delete the DB instance if both of the
    following conditions are true:   The DB cluster is a read replica of another Amazon Aurora DB cluster.   The DB
    instance is the only instance in the DB cluster.   To delete a DB instance in this case, first call the
    PromoteReadReplicaDBCluster API action to promote the DB cluster so it's no longer a read replica. After the
    promotion completes, then call the DeleteDBInstance API action to delete the final instance in the DB cluster.

    Args:
        ctx:
        hub:
        name(Text): Idem name to identify the resource.
        resource_id(Text): AWS ID to identify the resource.
        skip_final_snapshot(bool, optional): Mention this true if you want to skip creating snapshot default is false.
                                             Either this or final_db_snapshot_identifier should be provided
        final_db_snapshot_identifier(Text, optional): Id for the created final db_snapshot while deleting
        timeout(Dict, optional): Timeout configuration for deletion of AWS DB Instance.
            * delete (Dict) -- Timeout configuration for deletion of a DB Instance
                * delay(int) -- The amount of time in seconds to wait between attempts.Defaults to 30
                * max_attempts(int) -- Customized timeout configuration containing delay and max attempts.Defaults to 60

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            test-db-instance:
              aws.rds.db_instance.absent:
                - resource_id: test-db-instance,
                - skip_final_snapshot: true
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)

    before = await hub.exec.boto3.client.rds.describe_db_instances(
        ctx, DBInstanceIdentifier=resource_id
    )

    if not before["ret"]:
        result["comment"] = (f"aws.rds.db_instance '{name}' already absent",)

    else:
        db_instance_arn = before["ret"]["DBInstances"][0].get("DBInstanceArn")
        ret_tags = await hub.exec.boto3.client.rds.list_tags_for_resource(
            ctx, ResourceName=db_instance_arn
        )
        if not ret_tags["result"]:
            result["comment"] = ret_tags["comment"]
            result["result"] = False
            return result
        result[
            "old_state"
        ] = await hub.tool.aws.rds.conversion_utils.convert_raw_db_instance_to_present(
            raw_resource=before["ret"]["DBInstances"][0],
            raw_resource_tags=ret_tags,
            idem_resource_name=name,
        )
        if ctx.get("test", False):
            result["comment"] = (f"Would delete aws.rds.db_instance '{name}'",)
            return result

        ret = await hub.exec.boto3.client.rds.delete_db_instance(
            ctx,
            DBInstanceIdentifier=resource_id,
            SkipFinalSnapshot=skip_final_snapshot,
            FinalDBSnapshotIdentifier=final_db_snapshot_identifier,
        )
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            result["result"] = False
            return result
        # Waiting for the DBInstance to Delete
        waiter_config = hub.tool.aws.waiter_utils.create_waiter_config(
            default_delay=30,
            default_max_attempts=60,
            timeout_config=timeout.get("delete") if timeout else None,
        )
        try:
            await hub.tool.boto3.client.wait(
                ctx,
                "rds",
                "db_instance_deleted",
                DBInstanceIdentifier=resource_id,
                WaiterConfig=waiter_config,
            )
        except Exception as e:
            result["comment"] = (str(e),)
            result["result"] = False
        result["comment"] = result["comment"] + (
            f"Deleted aws.rds.db_instance '{name}'",
        )

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""
    **Autogenerated function**

    Describe the resource in a way that can be recreated/managed with the corresponding "present" function


    Returns information about provisioned RDS instances. This API supports pagination.  This operation can also
    return information for Amazon Neptune DB instances and Amazon DocumentDB instances.


    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: bash

            $ idem describe aws.rds.db_instance
    """

    result = {}
    ret = await hub.exec.boto3.client.rds.describe_db_instances(ctx)
    if not ret["result"]:
        hub.log.debug(f"Could not describe db_instance {ret['comment']}")
        return result
    for db_instance in ret["ret"]["DBInstances"]:
        # Including fields to match the 'present' function parameters
        resource_id = db_instance.get("DBInstanceIdentifier")
        db_instance_arn = db_instance.get("DBInstanceArn")
        ret_tags = await hub.exec.boto3.client.rds.list_tags_for_resource(
            ctx, ResourceName=db_instance_arn
        )
        resource_translated = (
            await hub.tool.aws.rds.conversion_utils.convert_raw_db_instance_to_present(
                raw_resource=db_instance,
                raw_resource_tags=ret_tags,
                idem_resource_name=resource_id,
            )
        )
        result[resource_id] = {
            "aws.rds.db_instance.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource_translated.items()
            ]
        }

    return result



"""
Autogenerated using `pop-create-idem <https://gitlab.com/saltstack/pop/pop-create-idem>`__

hub.exec.boto3.client.ec2.create_transit_gateway
hub.exec.boto3.client.ec2.delete_transit_gateway
hub.exec.boto3.client.ec2.describe_transit_gateways
hub.exec.boto3.client.ec2.modify_transit_gateway
"""
import copy
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]
__reconcile_wait__ = {"static": {"wait_in_seconds": 60}}
TREQ = {
    "present": {
        "require": ["aws.ec2.subnet.present", "aws.ec2.transit_gateway.present"],
    },
}


async def present(
    hub,
    ctx,
    name: str,
    transit_gateway: str,
    vpc_id: str,
    subnet_ids: List,
    resource_id: str = None,
    options: make_dataclass(
        "CreateTransitGatewayVpcAttachmentRequestOptions",
        [
            ("DnsSupport", str, field(default=None)),
            ("Ipv6Support", str, field(default=None)),
            ("ApplianceModeSupport", str, field(default=None)),
        ],
    ) = None,
    tags: Dict[str, Any]
    or List[
        make_dataclass(
            "Tag",
            [("Key", str, field(default=None)), ("Value", str, field(default=None))],
        )
    ] = None,
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Attaches the specified VPC to the specified transit gateway. If you attach a VPC with a CIDR range that overlaps
    the CIDR range of a VPC that is already attached, the new VPC CIDR range is not propagated to the default
    propagation route table. To send VPC traffic to an attached transit gateway, add a route to the VPC route table
    using CreateRoute.

    Args:
        name(Text): Name of the resource.
        transit_gateway(Text): The ID of the transit gateway.
        vpc_id(Text): The ID of the VPC.
        subnet_ids(List): The IDs of one or more subnets. You can specify only one subnet per Availability Zone.
                       You must specify at least one subnet, but we recommend that you specify two subnets for better availability.
                       The transit gateway uses one IP address from each specified subnet.
             * (Text)
        resource_id(Text, optional): AWS Transit gateway VPC attachment id.
        options(Dict, optional): The VPC attachment options. Defaults to None.
            * (DnsSupport, optional): Enable or disable DNS support. The default is enable.
            * (Ipv6Support, optional): Enable or disable IPv6 support. The default is disable.
            * (ApplianceModeSupport, optional): Enable or disable support for appliance mode. If enabled, a traffic flow between a source and
                destination uses the same Availability Zone for the VPC attachment for the lifetime of that
                flow. The default is disable.
        tags(Dict or List, optional): Dict in the format of {tag-key: tag-value} or List of tags in the format of
            [{"Key": tag-key, "Value": tag-value}] to associate with the transit gateway VPC attachment.
            Each tag consists of a key name and an associated value. Defaults to None.
            * (Key, optional): The key of the tag. Constraints: Tag keys are case-sensitive and accept a maximum of 127 Unicode
                characters. May not begin with aws:.
            * (Value, optional): The value of the tag. Constraints: Tag values are case-sensitive and accept a maximum of 256
                Unicode characters.

    Request Syntax:
        [transit-gateway-resource-id]:
          aws.ec2.transit_gateway_vpc_attachment.present:
            - name: 'string'
            - resource_id: 'string'
            - transit_gateway: 'string'
            - vpc_id: 'string'
            - subnet_ids:
               - 'string'
            - options:
                ApplianceModeSupport: 'string'
                DnsSupport: 'string'
                Ipv6Support: 'string'
            - tags:
                - Key: 'string'
                  Value: 'string'
    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            tgw-attach-0871e9c72becf0710:
                aws.ec2.transit_gateway_vpc_attachment.present:
                - name: tgw-attach-0871e9c72becf0710
                - resource_id: tgw-attach-0871e9c72becf0710
                - transit_gateway: tgw-02994a8dda824c337
                - vpc_id: vpc-0afa0d5fe3fc2785c
                - subnet_ids:
                   - subnet-07f91b9ebd252be49
                - options:
                    ApplianceModeSupport: disable
                    DnsSupport: enable
                    Ipv6Support: disable
                - tags:
                    - Key: Name
                      Value: test-transit-gateway-attachment

    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    before = None
    resource_updated = False
    if resource_id:
        before = await hub.exec.aws.ec2.transit_gateway_vpc_attachment.get_transit_gateway_vpc_attachment_by_id(
            ctx, resource_id
        )

    if isinstance(tags, List):
        tags = hub.tool.aws.tag_utils.convert_tag_list_to_dict(tags)

    if before and before["ret"]:
        result[
            "old_state"
        ] = hub.tool.aws.ec2.conversion_utils.convert_raw_tg_vpc_attachment(
            before["ret"], name
        )

    try:
        if before and before["ret"]:
            plan_state = copy.deepcopy(result["old_state"])
            current_status = result["old_state"]["state"]
            if current_status == "available":
                if subnet_ids is not None:
                    update_ret = await hub.exec.aws.ec2.transit_gateway_vpc_attachment.update_transit_gateway_vpc_attachment(
                        ctx=ctx,
                        transit_gateway_vpc_attachment_id=resource_id,
                        old_subnets=before["ret"].get("SubnetIds"),
                        new_subnets=subnet_ids,
                        old_options=before["ret"].get("Options"),
                        new_options=options,
                    )
                    result["comment"] = result["comment"] + update_ret["comment"]
                    result["result"] = update_ret["result"]
                    resource_updated = resource_updated or update_ret["ret"]
                    if ctx.get("test", False) and update_ret["ret"]:
                        if subnet_ids:
                            plan_state["subnet_ids"] = update_ret["ret"].get(
                                "subnet_ids"
                            )
                        if options:
                            plan_state["options"] = update_ret["ret"].get("options")
                if result["result"] and (
                    tags is not None and tags != result["old_state"].get("tags")
                ):
                    # Update tags
                    update_ret = await hub.exec.aws.ec2.tag.update_tags(
                        ctx=ctx,
                        resource_id=resource_id,
                        old_tags=before.get("Tags"),
                        new_tags=tags,
                    )
                    if not update_ret["result"]:
                        result["comment"] = result["comment"] + update_ret["comment"]
                        result["result"] = update_ret["result"]
                    resource_updated = resource_updated or update_ret["result"]
                    if ctx.get("test", False) and update_ret["result"]:
                        plan_state["tags"] = update_ret["ret"]
                if resource_updated:
                    result["comment"] = result["comment"] + (
                        f"'Updated aws.ec2.transit_gateway_vpc_attachment {name}'",
                    )
                else:
                    result["comment"] = result["comment"] + (f"no changes to update",)
            else:
                result["new_state"] = copy.deepcopy(result["old_state"])
                result["comment"] = result["comment"] + (
                    f"aws.ec2.transit_gateway_vpc_attachment {name} is in {current_status} state",
                )

        else:
            if ctx.get("test", False):
                result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                    enforced_state={},
                    desired_state={
                        "name": name,
                        "transit_gateway": transit_gateway,
                        "vpc_id": vpc_id,
                        "subnet_ids": subnet_ids,
                        "options": options,
                        "tags": tags,
                    },
                )
                result["comment"] = result["comment"] + (
                    f"Would create aws.ec2.transit_gateway_vpc_attachment {name}",
                )
                return result

            ret = await hub.exec.boto3.client.ec2.create_transit_gateway_vpc_attachment(
                ctx,
                TransitGatewayId=transit_gateway,
                VpcId=vpc_id,
                SubnetIds=subnet_ids,
                Options=options,
                TagSpecifications=[
                    {
                        "ResourceType": "transit-gateway-attachment",
                        "Tags": hub.tool.aws.tag_utils.convert_tag_dict_to_list(tags),
                    }
                ]
                if tags
                else None,
            )

            if not ret["result"]:
                result["comment"] = result["comment"] + ret["comment"]
                result["result"] = ret["result"]
                return result

            result["comment"] = result["comment"] + (
                f"Created aws.ec2.transit_gateway_vpc_attachment '{name}'",
            )
            resource_updated = resource_updated or ret["result"]
            resource_id = ret["ret"]["TransitGatewayVpcAttachment"][
                "TransitGatewayAttachmentId"
            ]

    except hub.tool.boto3.exception.ClientError as e:
        result["result"] = False
        result["comment"] = result["comment"] + (f"{e.__class__.__name__}: {e}",)

    try:
        if ctx.get("test", False):
            result["new_state"] = plan_state
        elif resource_updated:
            after = await hub.exec.aws.ec2.transit_gateway_vpc_attachment.get_transit_gateway_vpc_attachment_by_id(
                ctx, resource_id
            )
            result[
                "new_state"
            ] = hub.tool.aws.ec2.conversion_utils.convert_raw_tg_vpc_attachment(
                after["ret"], name
            )
        else:
            result["new_state"] = copy.deepcopy(result["old_state"])
    except Exception as e:
        result["comment"] = result["comment"] + (str(e),)
        result["result"] = False
    return result


async def absent(hub, ctx, name: str, resource_id: str = None) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Deletes the specified VPC attachment.

    Args:
        name(Text): Name of the resource.
        resource_id(Text, optional): AWS Transit gateway VPC attachment id. Idem automatically considers this resource being absent
         if this field is not specified.

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            aws.ec2.transit_gateway_vpc_attachment.absent:
                - name: tgw-attach-0871e9c72becf0710
                - resource_id: tgw-attach-0871e9c72becf0710
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    if not resource_id:
        result["comment"] = (
            f"aws.ec2.transit_gateway_vpc_attachment '{name}' already absent",
        )
        return result
    before = await hub.exec.aws.ec2.transit_gateway_vpc_attachment.get_transit_gateway_vpc_attachment_by_id(
        ctx, resource_id
    )
    if not before or not before["ret"]:
        result["comment"] = (
            f"aws.ec2.transit_gateway_vpc_attachment '{name}' already absent",
        )
        return result
    before = before["ret"]
    if not before or (before["State"] and (before["State"] == "deleted")):
        result["comment"] = result["comment"] + (
            f"aws.ec2.transit_gateway_vpc_attachment '{name}' is in deleted state.",
        )
    elif ctx.get("test", False):
        result[
            "old_state"
        ] = hub.tool.aws.ec2.conversion_utils.convert_raw_tg_vpc_attachment(
            before, name
        )
        result["comment"] = result["comment"] + (
            f"Would delete aws.ec2.transit_gateway_vpc_attachment {name}",
        )
        return result
    else:
        result[
            "old_state"
        ] = hub.tool.aws.ec2.conversion_utils.convert_raw_tg_vpc_attachment(
            before, name
        )
        try:
            ret = await hub.exec.boto3.client.ec2.delete_transit_gateway_vpc_attachment(
                ctx, TransitGatewayAttachmentId=resource_id
            )
            if not ret["result"]:
                result["comment"] = result["comment"] + ret["comment"]
                result["result"] = ret["result"]
                return result
            after = await hub.exec.aws.ec2.transit_gateway_vpc_attachment.get_transit_gateway_vpc_attachment_by_id(
                ctx, resource_id
            )
            result[
                "new_state"
            ] = hub.tool.aws.ec2.conversion_utils.convert_raw_tg_vpc_attachment(
                after["ret"], name
            )
            result["comment"] = result["comment"] + (
                f"Deleted aws.ec2.transit_gateway_vpc_attachment '{name}'",
            )
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = result["comment"] + (f"{e.__class__.__name__}: {e}",)
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""
    **Autogenerated function**

    Describe the resource in a way that can be recreated/managed with the corresponding "present" function


    Describes one or more VPC attachments. By default, all VPC attachments are described. Alternatively, you can
    filter the results.


    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: bash

            $ idem describe aws.ec2.transit_gateway_vpc_attachment
    """

    result = {}

    ret = await hub.exec.boto3.client.ec2.describe_transit_gateway_vpc_attachments(ctx)
    if not ret["result"]:
        hub.log.debug(
            f"Could not describe aws.ec2.transit_gateway_vpc_attachment {ret['comment']}"
        )
        return {}
    for transit_gateway_vpc_attachment in ret["ret"]["TransitGatewayVpcAttachments"]:
        resource_id = transit_gateway_vpc_attachment.get("TransitGatewayAttachmentId")
        translated_resource = (
            hub.tool.aws.ec2.conversion_utils.convert_raw_tg_vpc_attachment(
                transit_gateway_vpc_attachment, resource_id
            )
        )
        result[resource_id] = {
            "aws.ec2.transit_gateway_vpc_attachment.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in translated_resource.items()
            ]
        }

    return result


def is_pending(hub, ret):
    new_state = ret.get("new_state", None)
    status = None
    resource_id = None
    if new_state:
        status = new_state.get("state")
        resource_id = new_state.get("resource_id")
    if status and isinstance(status, str):
        hub.log.debug(
            f"Transit gateway VPC attachment: {resource_id} is_pending() status {status}"
        )
        if (
            status.casefold() == "available"
            or status.casefold() == "deleted"
            or status.casefold() == "failed"
            or status.casefold() == "rejected"
        ):
            hub.log.debug(
                f"No need to reconcile new state {new_state} for Transit gateway VPC attachment: {resource_id} with status {status}"
            )
            return False
        if (
            status.casefold() == "initiatingRequest"
            or status.casefold() == "pendingAcceptance"
            or status.casefold() == "rollingBack"
            or status.casefold() == "pending"
            or status.casefold() == "modifying"
            or status.casefold() == "deleting"
            or status.casefold() == "rejecting"
            or status.casefold() == "failing"
        ):
            hub.log.debug(
                f"Reconcile new state {new_state} for Transit gateway VPC attachment: {resource_id} with status {status}"
            )
            return True
    return (not ret["result"]) or bool(ret["changes"])



from typing import Dict
from typing import List

from dict_tools import data


def get_notification_configuration_diff(
    hub, new_notifications: Dict[str, List], old_state: Dict[str, List]
) -> Dict[str, List]:
    """
    This functions helps in comparing two dicts.
    It compares each key value in both the dicts and return diff notifications configuration

    Returns:
        {Resultant notifications dictionary}

    """

    old_state_notification = old_state.get("notifications")

    old_state_notification = (
        old_state_notification if old_state_notification is not None else {}
    )

    return data.recursive_diff(
        new_notifications, old_state_notification, ignore_order=True
    )



import copy
from typing import Any
from typing import Dict

__contracts__ = ["resource"]

TREQ = {
    "present": {
        "require": [
            "aws.sns.topic.present",
        ],
    },
}


async def present(
    hub,
    ctx,
    name: str,
    topic_arn: str,
    protocol: str,
    endpoint: str,
    resource_id: str = None,
    attributes: Dict = None,
    return_subscription_arn: bool = None,
) -> Dict[str, Any]:
    r"""

    Creates a new endpoint subscription to a topic.If the endpoint type is HTTP/S or email, or if the endpoint and
    the topic are not in the same Amazon Web Services account,For http and email protocols the endpoint owner must
    run the ConfirmSubscription action to confirm the subscription.

    Args:
        hub:
        ctx:
        name(Text): The idem name for the resource
        topic_arn(Text): The ARN of the topic to which the subscription should be added
        protocol(Text): The protocol that you want to use.Supported protocols are:http,https,email,email-json,
                                  sms,sqs,application,lambda,firehose
        endpoint(Text): The endpoint that you want to receive notifications.Endpoints vary by protocol
        resource_id(Text, optional): The AWS resource identifier, here it is resource arn
        attributes(Dict, optional): Attributes of the subscription
        return_subscription_arn(bool, optional): A bool value to specify if the subscription_arn should be return after
                                                creation.By default,it is false i.e. subscription_arn is not returned.


    Request Syntax:
        [subscription-name]:
          aws.sns.subscription.present:
          - topic_arn: 'string'
          - protocol: 'string'
          - endpoint: 'string'
          - resource_id: 'string'
          - attributes: 'dict'
          - return_subscription_arn: 'bool'


    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            subscription-name:
              aws.sns.subscription.present:
                - topic_arn: arn:aws:sns:eu-west-3:537227425989:test-topic
                - protocol: sms
                - endpoint: "+911234567890"
                - attributes:
                    DeliveryPolicy: '{"healthyRetryPolicy": {"minDelayTarget": 10,"maxDelayTarget": 30,"numRetries": 10,"numMaxDelayRetries": 7,"numNoDelayRetries": 0,"numMinDelayRetries": 3,"backoffFunction": "linear"},"sicklyRetryPolicy": null,"throttlePolicy": null,"guaranteed": false}'
                    FilterPolicy: '{"pets": ["dog", "cat"]}'
                - return_subscription_arn: True
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    before = None
    resource_updated = False
    plan_state = None
    # Standardise JSON string format
    if attributes:
        for key, value in attributes.items():
            attributes[key] = hub.tool.aws.sns.conversion_utils.standardise_json(value)

    if resource_id:
        resource = hub.tool.boto3.resource.create(
            ctx, "sns", "Subscription", arn=resource_id
        )
        before = await hub.tool.boto3.resource.describe(resource)

    if before:
        try:
            result[
                "old_state"
            ] = hub.tool.aws.sns.conversion_utils.convert_raw_subscription_to_present(
                raw_resource=before, idem_resource_name=name
            )

            plan_state = copy.deepcopy(result["old_state"])
            old_attributes = plan_state.get("attributes")

            if attributes is not None:
                # Update Attributes
                update_attributes_ret = (
                    await hub.exec.aws.sns.attributes.update_subscription_attributes(
                        ctx=ctx,
                        resource_arn=resource_id,
                        old_attributes=old_attributes,
                        new_attributes=attributes,
                    )
                )
                if not update_attributes_ret["result"]:
                    result["comment"] = update_attributes_ret["comment"]
                    result["result"] = False
                    return result
                resource_updated = resource_updated or bool(
                    update_attributes_ret["ret"]
                )
                result["comment"] = result["comment"] + update_attributes_ret["comment"]

                if ctx.get("test", False) and update_attributes_ret["ret"]:
                    plan_state["attributes"] = update_attributes_ret["ret"].get(
                        "updated_attributes"
                    )

            else:
                result["comment"] = (f"aws.sns.subscription '{name}' already exists",)

        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = result["comment"] + (f"{e.__class__.__name__}: {e}",)
            result["result"] = False

    else:
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "topic_arn": topic_arn,
                    "protocol": protocol,
                    "endpoint": endpoint,
                    "attributes": attributes,
                    "return_subscription_arn": return_subscription_arn,
                },
            )
            result["comment"] = (f"Would create aws.sns.subscription '{name}'",)
            return result
        try:
            ret = await hub.exec.boto3.client.sns.subscribe(
                ctx,
                TopicArn=topic_arn,
                Protocol=protocol,
                Endpoint=endpoint,
                Attributes=attributes,
                ReturnSubscriptionArn=True,
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                return result
            resource_id = ret["ret"].get("SubscriptionArn")
            result["comment"] = (f"Created aws.sns.subscription '{name}'",)
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = result["comment"] + (f"{e.__class__.__name__}: {e}",)
            result["result"] = False

    try:
        if ctx.get("test", False):
            result["new_state"] = plan_state
        elif (not before) or resource_updated:
            resource = hub.tool.boto3.resource.create(
                ctx, "sns", "Subscription", arn=resource_id
            )
            after = await hub.tool.boto3.resource.describe(resource)
            result[
                "new_state"
            ] = hub.tool.aws.sns.conversion_utils.convert_raw_subscription_to_present(
                raw_resource=after, idem_resource_name=name
            )
        else:
            result["new_state"] = copy.deepcopy(result["old_state"])
    except Exception as e:
        result["comment"] = result["comment"] + (str(e),)
        result["result"] = False
    return result


async def absent(hub, ctx, name: str, resource_id: str) -> Dict[str, Any]:
    r"""
    Deletes the specified subscription. This action is idempotent, so deleting a topic that does not exist
    does not result in an error.

    Args:
        name(Text): The name of the state.
        resource_id(Text): The AWS resource identifier i.e. resource arn

    Request Syntax:
        [subscription-name]:
          aws.sns.subscription.absent:
          - resource_id: 'string'

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            test-subscription:
              aws.sns.subscription.absent:
              - resource_id: arn:aws:sns:eu-west-3:537227425989:test-topic:9c0fd640-7fc6-4888-bc08-f0a497a6237f
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    before = None

    # There is a delay with describe() call, to check if the resource is in list_subscriptions() before describe()
    ret = await hub.exec.boto3.client.sns.list_subscriptions(ctx)
    if not ret["result"]:
        result["result"] = False
        result["comment"] = ret["comment"]
        return result
    for subscription in ret["ret"]["Subscriptions"]:
        if subscription.get("SubscriptionArn") == resource_id:
            resource = hub.tool.boto3.resource.create(
                ctx, "sns", "Subscription", arn=resource_id
            )
            before = await hub.tool.boto3.resource.describe(resource)
            break

    if not before:
        result["comment"] = (f"aws.sns.subscription '{name}' already absent",)

    else:
        result[
            "old_state"
        ] = hub.tool.aws.sns.conversion_utils.convert_raw_subscription_to_present(
            raw_resource=before, idem_resource_name=name
        )
        if ctx.get("test", False):
            result["comment"] = (f"Would delete aws.sns.subscription '{name}'",)
            return result
        try:
            ret = await hub.exec.boto3.client.sns.unsubscribe(
                ctx, SubscriptionArn=resource_id
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                result["result"] = False
                return result
            result["comment"] = (f"Deleted aws.sns.subscription '{name}'",)
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = result["comment"] + (f"{e.__class__.__name__}: {e}",)

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""
    Describe the resource in a way that can be recreated/managed with the corresponding "present" function
    Describes list of all the subscriptions present in all topics


    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: bash

            $ idem describe aws.sns.subscription
    """

    result = {}
    ret = await hub.exec.boto3.client.sns.list_subscriptions(ctx)
    if not ret["result"]:
        hub.log.debug(f"Could not describe subscriptions {ret['comment']}")
        return {}

    for subscription in ret["ret"]["Subscriptions"]:
        # Including fields to match the 'present' function parameters
        resource_id = subscription.get("SubscriptionArn")
        # Few deleted subscriptions might take 2-3 days to be removed from the list.
        # And few subscriptions might not be confirmed by endpoint yet.
        if resource_id == "Deleted" or resource_id == "PendingConfirmation":
            continue
        # Attributes are not present in the return value of list_subscriptions
        resource = hub.tool.boto3.resource.create(
            ctx, "sns", "Subscription", arn=resource_id
        )
        raw_resource = await hub.tool.boto3.resource.describe(resource)
        resource_translated = (
            hub.tool.aws.sns.conversion_utils.convert_raw_subscription_to_present(
                raw_resource=raw_resource, idem_resource_name=resource_id
            )
        )
        result[resource_id] = {
            "aws.sns.subscription.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource_translated.items()
            ]
        }
    return result



from typing import Any
from typing import Dict
from typing import List

import boto3.resources.factory


async def update_attachments(
    hub,
    ctx,
    internet_gateway_name: str,
    before: boto3.resources.factory.ServiceResource.meta,
    vpc_id: List = None,
) -> Dict[str, Any]:
    """
    Updates attachments of internet gateway. 1. vpc_id is none then there will be no change in attachments. 2. vpc_id
    is empty then internet gateway will be detached if it is already attached to a vpc 3. vpc_id has one element then
    first detach will be performed if already attached to another vpc and the attach to new vpc. If attached vpc and
    passed in vpc are same then there will be no update to the attachments

    Args:
        hub:
        ctx:
        internet_gateway_name: Internet gateway name
        vpc_id: This list can contain only single element. This is ID of VPC to which internet gateway attaches.
        before: Contains current configuration for the resource

    Returns:
        {"result": True|False, "comment": "A message", "ret": Dict}

    """
    result = dict(comment=(), result=True, ret=None)
    already_attached_vpc_id = (
        before["attachments"][0]["VpcId"] if "attachments" in before else None
    )
    if vpc_id is None:
        result["ret"] = {"updated_vpc_id": [already_attached_vpc_id]}
        result["comment"] = (f"'No change in attachments for {internet_gateway_name}'",)
        return result
    updated_vpc_id = []
    if ctx.get("test", False):
        if len(vpc_id) > 0:
            updated_vpc_id = [vpc_id[0]]
    else:
        if vpc_id:
            # Attach to new vpc. First, detach if already attached to another vpc and then perform attach.
            if vpc_id[0] == already_attached_vpc_id:
                result["comment"] = (
                    f"'{internet_gateway_name}' is already attached to vpc '{already_attached_vpc_id}'",
                )
            else:
                try:
                    if already_attached_vpc_id:
                        ret = await hub.exec.boto3.client.ec2.detach_internet_gateway(
                            ctx,
                            InternetGatewayId=internet_gateway_name,
                            VpcId=already_attached_vpc_id,
                        )
                        result["result"] = ret["result"]
                        if not result["result"]:
                            result["comment"] = ret["comment"]
                            result["result"] = False
                            return result
                        result["comment"] = (
                            f"'{internet_gateway_name}' detached from vpc '{already_attached_vpc_id}'",
                        )
                    ret = await hub.exec.boto3.client.ec2.attach_internet_gateway(
                        ctx, InternetGatewayId=internet_gateway_name, VpcId=vpc_id[0]
                    )
                    result["result"] = ret["result"]
                    if not result["result"]:
                        result["comment"] = ret["comment"]
                        result["result"] = False
                        return result
                    result["comment"] = result["comment"] + (
                        f"'{internet_gateway_name}' attached to vpc '{vpc_id[0]}'",
                    )
                except hub.tool.boto3.exception.ClientError as e:
                    result["comment"] = (
                        result["comment"] + f"{e.__class__.__name__}: {e}"
                    )
                    result["result"] = False
        else:
            try:
                # When vpc_id is empty, detach from vpc if already attached
                if "attachments" in before:
                    ret = await hub.exec.boto3.client.ec2.detach_internet_gateway(
                        ctx,
                        InternetGatewayId=internet_gateway_name,
                        VpcId=already_attached_vpc_id,
                    )
                    result["result"] = ret["result"]
                    if not result["result"]:
                        result["comment"] = ret["comment"]
                        result["result"] = False
                        return result
                    result["comment"] = (
                        f"'{internet_gateway_name}' detached from vpc '{already_attached_vpc_id}'",
                    )
                else:
                    result["comment"] = (
                        f"'{internet_gateway_name}' is already detached'",
                    )
            except hub.tool.boto3.exception.ClientError as e:
                result["comment"] = f"{e.__class__.__name__}: {e}"
                result["result"] = False
    result["ret"] = {"updated_vpc_id": updated_vpc_id}
    return result



"""
Autogenerated using `pop-create-idem <https://gitlab.com/saltstack/pop/pop-create-idem>`__

hub.exec.boto3.client.organizations.attach_policy
hub.exec.boto3.client.organizations.create_policy
hub.exec.boto3.client.organizations.delete_policy
hub.exec.boto3.client.organizations.describe_policy
hub.exec.boto3.client.organizations.detach_policy
hub.exec.boto3.client.organizations.list_policies
hub.exec.boto3.client.organizations.update_policy
"""
import copy
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]

TREQ = {
    "absent": {
        "require": [
            "aws.organizations.policy_attachment.absent",
        ],
    },
    "present": {
        "require": [
            "aws.organizations.account.present",
        ],
    },
}


async def present(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
    description: str = None,
    policy_type: str = None,
    content: str = None,
    tags: List = None,
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Creates a policy of a specified type that you can attach to a root, an organizational unit (OU), or an
    individual AWS account. For more information about policies and their use, see Managing Organization Policies.
    If the request includes tags, then the requester must have the organizations:TagResource permission. This
    operation can be called only from the organization's management account.

    Args:
        name(Text): Name of the resource.
        resource_id(Text, Optional): AWS Organization policy ID to identify the resource.
        description: An optional description to assign to the policy.
        policy_type : The type of policy to create. Only supported type currently is SERVICE_CONTROL_POLICY
        content : The policy text content to add to the new policy. The text that you supply must adhere to
                    the rules of the policy type you specify in the Type parameter
        tags(Optional): A list of tags that you want to attach to the newly created policy

    Request Syntax:
        [policy-name]:
          aws.organizations.policy.present:
          - resource_id: 'string'
          - name: 'string'
          - description: 'string'
          - policy_type: 'string'
          - content: 'string'
          - tags:
            - Key: 'string'
              Value: 'string'

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            p-name:
              aws.organizations.policy.present:
                - resource_id: p_id
                - name: AllowAllS3Actions
                - description: Enables admins of attached accounts to delegate all S3 permissions
                - policy_type : SERVICE_CONTROL_POLICY
                - content : '{\"Version\":\"2012-10-17\",\"Statement\":{\"Effect\":\"Allow\",\"Action\":\"s3:*\"}}'
                - tags:
                  - Key: org
                    Value: idem-aws
    """

    result = dict(comment=(), name=name, result=True, old_state=None, new_state=None)
    before = None
    existing_tags = None

    if resource_id:
        before = await hub.exec.boto3.client.organizations.describe_policy(
            ctx, PolicyId=resource_id
        )

    change_dict = dict(Name=None, Description=None, Content=None)

    update_tag = False
    policy_id = None
    if before:
        # Policy exists , update
        if not ctx.get("test", False):
            result["comment"] = result["comment"] + (
                f"aws.organizations.policy '{name}' exists.",
            )
        result[
            "old_state"
        ] = hub.tool.aws.organization.conversion_utils.convert_raw_policy_to_present(
            before["ret"]["Policy"]
        )
        plan_state = copy.deepcopy(result["old_state"])

        policy_id = before["ret"]["Policy"]["PolicySummary"].get("Id")

        try:

            if before["ret"]["Policy"]["PolicySummary"].get("Name") != name:
                change_dict["Name"] = name

            if (
                before["ret"]["Policy"]["PolicySummary"].get("Description")
                != description
            ):
                change_dict["Description"] = description

            if before["ret"]["Policy"]["Content"] != content:
                change_dict["Content"] = content

            if any(value is not None for value in change_dict.values()):

                if ctx.get("test", False):
                    plan_state["resource_id"] = policy_id
                    if change_dict.get("Name"):
                        plan_state["name"] = change_dict.get("Name")
                    if change_dict.get("Description"):
                        plan_state["description"] = change_dict.get("Description")
                    if change_dict.get("Content"):
                        plan_state["content"] = change_dict.get("Content")
                    result["comment"] = result["comment"] + (
                        f"Would update aws.organizations.policy '{name}'",
                    )
                else:
                    update_policy_ret = (
                        await hub.exec.boto3.client.organizations.update_policy(
                            ctx,
                            PolicyId=policy_id,
                            Name=change_dict.get("Name"),
                            Description=change_dict.get("Description"),
                            Content=change_dict.get("Content"),
                        )
                    )

                    if not update_policy_ret["result"]:
                        result["comment"] = (
                            result["comment"] + update_policy_ret["comment"]
                        )
                        result["result"] = update_policy_ret["result"]
                        return result
                    after = update_policy_ret

                    if after["ret"]:
                        result[
                            "new_state"
                        ] = hub.tool.aws.organization.conversion_utils.convert_raw_policy_to_present(
                            after["ret"]["Policy"]
                        )
                        result["comment"] = result["comment"] + (
                            f"Updated aws.organizations.policy '{name}'",
                        )

            else:
                result["new_state"] = copy.deepcopy(result["old_state"])

            old_tags = await hub.exec.boto3.client.organizations.list_tags_for_resource(
                ctx, ResourceId=policy_id
            )

            if old_tags["result"]:
                existing_tags = old_tags["ret"]["Tags"]
                result["old_state"]["tags"] = old_tags["ret"]["Tags"]

                # If no update in tags , then new_state will contain same tags as old
                if tags is None and not ctx.get("test", False):
                    result["new_state"]["tags"] = result["old_state"]["tags"]

            if tags is not None:

                update_tags_ret = (
                    await hub.exec.aws.organizations.organization.update_tags(
                        ctx, policy_id, existing_tags if existing_tags else [], tags
                    )
                )
                if not update_tags_ret["result"]:
                    result["comment"] = result["comment"] + update_tags_ret["comment"]
                    result["result"] = update_tags_ret["result"]
                    return result

                if update_tags_ret["ret"]:
                    update_tag = True
                    if ctx.get("test", False):
                        plan_state["tags"] = tags

                    result["comment"] = result["comment"] + (
                        "Updated tags on aws.organizations.policy '{name}'.",
                    )

        except hub.tool.boto3.exception.ClientError as e:
            result["result"] = False
            result["comment"] = result["comment"] + (f"{e.__class__.__name__}: {e}",)

    else:

        # Policy does not exist , create

        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "description": description,
                    "policy_type": policy_type,
                    "content": content,
                    "tags": tags,
                },
            )
            result["comment"] = result["comment"] + (
                f"Would create aws.organizations.policy '{name}'",
            )
            return result

        try:
            create_policy_ret = await hub.exec.boto3.client.organizations.create_policy(
                ctx,
                Name=name,
                Description=description,
                Type=policy_type,
                Content=content,
                Tags=tags,
            )
            result["result"] = create_policy_ret["result"]
            if not result["result"]:
                result["comment"] = result["comment"] + create_policy_ret["comment"]
                return result

            after = create_policy_ret

            result[
                "new_state"
            ] = hub.tool.aws.organization.conversion_utils.convert_raw_policy_to_present(
                after["ret"]["Policy"]
            )

            policy_id = result["new_state"]["resource_id"]
            if tags is not None:
                list_tags = (
                    await hub.exec.boto3.client.organizations.list_tags_for_resource(
                        ctx, ResourceId=policy_id
                    )
                )

                if list_tags["result"]:
                    result["new_state"]["tags"] = list_tags["ret"]["Tags"]
                else:
                    hub.log.debug(
                        f"Unable to list tags for resource {resource_id} with error: {list_tags['comment']}"
                    )
                    result["comment"] = result["comment"] + list_tags["comment"]
                    result["result"] = list_tags["result"]

            result["comment"] = result["comment"] + (
                f"Created aws.organizations.policy '{name}'.",
            )
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = result["comment"] + (f"{e.__class__.__name__}: {e}",)
            result["result"] = False

    if ctx.get("test", False):
        result["new_state"] = plan_state
    elif before and update_tag:
        try:
            updated_tags = (
                await hub.exec.boto3.client.organizations.list_tags_for_resource(
                    ctx, ResourceId=policy_id
                )
            )

            if updated_tags["result"]:
                result["new_state"]["tags"] = updated_tags["ret"]["Tags"]
            else:
                hub.log.debug(
                    f"Unable to list tags for resource {resource_id} with error: {updated_tags['comment']}"
                )
                result["comment"] = result["comment"] + updated_tags["comment"]
                result["result"] = updated_tags["result"]

        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = result["comment"] + (f"{e.__class__.__name__}: {e}",)
            result["result"] = False
    return result


async def absent(hub, ctx, name: str, resource_id: str) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Deletes the specified policy from your organization. Before you perform this operation, you must first detach
    the policy from all organizational units (OUs), roots, and accounts. This operation can be called only from the
    organization's management account.

    Args:
        name(Text): Name of the resource.
        resource_id(Text): WS Organization policy ID to identify the resource.

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls
            p-iauqv2gg:
                aws.organizations.policy.absent:
                - name: policy_name_1
                - resource_id: p-iauqv2gg
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)

    before = await hub.exec.boto3.client.organizations.describe_policy(
        ctx, PolicyId=resource_id
    )

    if not before:
        result["comment"] = result["comment"] + (
            f"aws.organizations.policy '{name}' already absent",
        )
    else:
        result[
            "old_state"
        ] = hub.tool.aws.organization.conversion_utils.convert_raw_policy_to_present(
            before["ret"]["Policy"]
        )
        if ctx.get("test", False):
            result["comment"] = result["comment"] + (
                f"Would delete aws.organizations.policy '{name}'",
            )
            return result
        try:

            delete_ret = await hub.exec.boto3.client.organizations.delete_policy(
                ctx, PolicyId=resource_id
            )

            if not delete_ret["result"]:
                result["comment"] = result["comment"] + delete_ret["comment"]
                result["result"] = delete_ret["result"]
                return result

            result["comment"] = result["comment"] + (
                f"aws.organizations.policy {name} deleted.",
            )
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = result["comment"] + (f"{e.__class__.__name__}: {e}",)
            result["result"] = False

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""
    **Autogenerated function**

    Retrieves information about a policy.

    This operation can be called only from the organization's management account
    or by a member account that is a delegated administrator for an AWS service.

    Currently, IDEM AWS supports only SERVICE_CONTROL_POLICY.

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: bash

            $ idem describe aws.organizations.policy
    """

    result = {}

    list_all_policies = await hub.exec.boto3.client.organizations.list_policies(
        ctx, Filter="SERVICE_CONTROL_POLICY"
    )

    if not list_all_policies:
        hub.log.debug(
            f"Could not describe policy with error: {list_all_policies['comment']}"
        )
        return {}

    for policy in list_all_policies["ret"]["Policies"]:
        policy_id = policy["Id"]

        policy_ret = await hub.exec.boto3.client.organizations.describe_policy(
            ctx, PolicyId=policy_id
        )

        if not policy_ret and not policy_ret["ret"]:
            hub.log.debug(
                f"Could not describe '{policy_id}' with error: {policy_ret['comment']}"
            )
            continue

        tags_ret = await hub.exec.boto3.client.organizations.list_tags_for_resource(
            ctx, ResourceId=policy_id
        )
        if not tags_ret["result"]:
            hub.log.debug(
                f"Unable to list tags for resource {policy_id} with error: {tags_ret['comment']}"
            )

        translated_resource = (
            hub.tool.aws.organization.conversion_utils.convert_raw_policy_to_present(
                policy_ret["ret"]["Policy"],
                tags_ret["ret"]["Tags"] if tags_ret else None,
            )
        )

        result[policy["Name"]] = {
            "aws.organizations.policy.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in translated_resource.items()
            ]
        }

    return result



from typing import Any
from typing import Dict

import dict_tools.data


async def modify(
    hub, name: str, raw_profile: Dict[str, Any], new_profile: Dict[str, Any]
):
    """
    Assume a role using the credentials in the raw_profile.
    Create a new profile with keys and ids based on the assumed role.
    If assuming the role fails, the new_profile is made completely unusable.
    """
    if "assume_role" not in raw_profile:
        return

    # Get new access credentials for the assumed role
    assume_role_obj = raw_profile.pop("assume_role", {})

    ctx = dict_tools.data.NamespaceDict(acct=new_profile, test=False)

    try:
        credentials = await hub.exec.aws.sts.assume_role.credentials(
            ctx, **assume_role_obj
        )
    except Exception as e:
        credentials = dict_tools.data.NamespaceDict(
            result=False, comment=f"Failed to assume role: {e.__class__.__name__}: {e}"
        )

    if not credentials.result:
        # Failed to assume the role, don't allow a half-baked profile to be used
        # TODO After acct 8.1.0 is in idem an error can be raised instead here
        new_profile.clear()
        hub.log.error(f"Failed to assume role: {credentials.comment}")
        return
    # Overwrite the credentials for the profile with the ones from the assumed role
    new_profile["aws_access_key_id"] = credentials.ret["AccessKeyId"]
    new_profile["aws_secret_access_key"] = credentials.ret["SecretAccessKey"]
    new_profile["aws_session_token"] = credentials.ret["SessionToken"]



from collections import OrderedDict
from typing import Any
from typing import Dict


def convert_raw_monitor_to_present(
    hub, ctx, raw_resource: Dict[str, Any], idem_resource_name: str
) -> Dict[str, Any]:
    resource_id = raw_resource.get("MonitorArn")
    resource_parameters = OrderedDict(
        {
            "MonitorSpecification": "monitor_specification",
            "MonitorDimension": "monitor_dimension",
            "MonitorType": "monitor_type",
            "MonitorName": "monitor_name",
        }
    )

    resource_translated = {
        "name": idem_resource_name,
        "resource_id": resource_id,
    }
    for parameter_raw, parameter_present in resource_parameters.items():
        if parameter_raw in raw_resource:
            resource_translated[parameter_present] = raw_resource.get(parameter_raw)

    return resource_translated


def convert_raw_subscription_to_present(
    hub, ctx, raw_resource: Dict[str, Any], idem_resource_name: str
) -> Dict[str, Any]:
    resource_id = raw_resource.get("SubscriptionArn")
    resource_parameters = OrderedDict(
        {
            "MonitorArnList": "monitor_arn_list",
            "Subscribers": "subscribers",
            "Threshold": "threshold",
            "Frequency": "frequency",
            "SubscriptionName": "subscription_name",
        }
    )

    resource_translated = {"name": idem_resource_name, "resource_id": resource_id}
    for parameter_raw, parameter_present in resource_parameters.items():
        if parameter_raw in raw_resource:
            resource_translated[parameter_present] = raw_resource.get(parameter_raw)

    return resource_translated


async def convert_raw_cost_category_to_present(
    hub, ctx, raw_resource: Dict[str, Any], idem_resource_name: str
) -> Dict[str, Any]:
    resource_id = raw_resource.get("CostCategoryArn")

    resource_parameters = OrderedDict(
        {
            "Name": "cost_category_name",
            "EffectiveStart": "effective_start",
            "EffectiveEnd": "effective_end",
            "RuleVersion": "rule_version",
            "Rules": "rules",
            "SplitChargeRules": "split_charge_rules",
            "ProcessingStatus": "processing_status",
            "DefaultValue": "default_value",
            "InheritedValue": "inherited_value",
        }
    )

    resource_translated = {
        "name": idem_resource_name,
        "resource_id": resource_id,
    }

    ret_tag = await hub.exec.boto3.client.ce.list_tags_for_resource(
        ctx, ResourceArn=resource_id
    )
    if ret_tag["result"]:
        resource_translated["tags"] = hub.tool.aws.tag_utils.convert_tag_list_to_dict(
            ret_tag.get("ret").get("ResourceTags")
        )

    for parameter_raw, parameter_present in resource_parameters.items():
        if parameter_raw in raw_resource:
            resource_translated[parameter_present] = raw_resource.get(parameter_raw)

    return resource_translated



import copy
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    role_arn: str,
    recording_group: make_dataclass(
        "ConfigurationRecorder",
        [
            ("allSupported", bool, field(default=None)),
            ("includeGlobalResourceTypes", bool, field(default=None)),
            ("resourceTypes", List[str], field(default=None)),
        ],
    ),
    recording: bool = False,
    resource_id: str = None,
) -> Dict[str, Any]:
    """
    Creates a new configuration recorder to record the selected resource configurations, please see AWS Config services
    Args:
        name(Text): The name of the recorder.
        role_arn (Text): Amazon Resource Name (ARN) of the IAM role used to describe the Amazon Web Services resources associated with the account.
        recording_group (Dict): Specifies the types of Amazon Web Services resources for which Config records configuration changes.
        recording (Bool, Optional): Specifies recording status of the configuration recorder. Default is False.
        resource_id (Text, Optional): The name of the recorder.

    Request syntax:
        [aws-config-recorder]:
          aws.config.config_recorder.present:
          - name: 'string'
          - resource_id: 'string'
          - role_arn 'string'
          - recording: 'string'
          - recording_group: 'dict'

    Returns:
         Dict[str, Any]

    Examples:
        .. code-block:: sls
            aws-config-recorder:
              aws.config.config_recorder.present:
                - name: 'config_recorder'
                - resource_id: 'config_recorder'
                - role_arn 'arn:aws:iam::012345678912:role/aws-service-role/config.amazonaws.com/AWSServiceRoleForConfig'
                - recording: true
                - recording_group:
                    allSupported: false
                    includeGlobalResourceTypes: false
                    resourceTypes:
                    - "AWS::ApiGateway::Stage"

    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    before = None
    resource_updated: bool = False
    only_recording_status_changed: bool = False

    if resource_id:
        resource = await hub.exec.boto3.client.config.describe_configuration_recorders(
            ctx,
            ConfigurationRecorderNames=[name],
        )
        if not resource["result"]:
            result["comment"] = resource["comment"]
            result["result"] = False
            return result
        else:
            before = resource["ret"]["ConfigurationRecorders"][0]

    if before:
        before_recording_status = (
            await hub.exec.boto3.client.config.describe_configuration_recorder_status(
                ctx,
                ConfigurationRecorderNames=[name],
            )
        )
        if not before_recording_status["result"]:
            result["comment"] = before_recording_status["comment"]
            result["result"] = False
            return result
        before["recording"] = before_recording_status["ret"][
            "ConfigurationRecordersStatus"
        ][0]["recording"]

        resource_translated = await hub.tool.aws.config.conversion_utils.convert_raw_config_recorder_to_present(
            ctx,
            raw_resource=before,
            recording=before_recording_status["ret"]["ConfigurationRecordersStatus"][0][
                "recording"
            ],
        )
        result["old_state"] = resource_translated
        resource_updated = await hub.tool.aws.config.config_utils.is_resource_updated(
            before=resource_translated,
            role_arn=role_arn,
            recording_group=recording_group,
        )
        if not resource_updated and resource_translated["recording"] == recording:
            result["comment"] = (f"aws.config.config_recorder '{name}' already exists",)
            result["new_state"] = copy.deepcopy(result["old_state"])
            return result

        if not resource_updated and resource_translated["recording"] != recording:
            only_recording_status_changed = True

    if ctx.get("test", False):
        result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
            enforced_state={},
            desired_state={
                "name": name,
                "role_arn": role_arn,
                "recording": recording,
                "resource_id": resource_id,
                "recording_group": recording_group,
            },
        )
        operation = (
            "update"
            if before
            and (resource_updated or resource_translated["recording"] != recording)
            else "create"
        )
        result["comment"] = (f"Would {operation} aws.config.config_recorder '{name}'",)
        return result

    if not only_recording_status_changed:
        update_ret = await hub.exec.boto3.client.config.put_configuration_recorder(
            ctx=ctx,
            ConfigurationRecorder={
                "name": name,
                "roleARN": role_arn,
                "recordingGroup": recording_group,
            },
        )
        if not update_ret["result"]:
            result["comment"] = update_ret["comment"]
            result["result"] = False
            return result

        after_ret = await hub.exec.boto3.client.config.describe_configuration_recorders(
            ctx,
            ConfigurationRecorderNames=[name],
        )
        if not after_ret["result"]:
            result["comment"] = after_ret["comment"]
            result["result"] = False
            return result

        result[
            "new_state"
        ] = await hub.tool.aws.config.conversion_utils.convert_raw_config_recorder_to_present(
            ctx,
            raw_resource=after_ret["ret"]["ConfigurationRecorders"][0],
            recording=None,
        )
        operation = "Created" if not before else "Updated"
        result["comment"] = (f"{operation} aws.config.config_recorder '{name}'",)
    else:
        result["new_state"] = copy.deepcopy(result["old_state"])

    # If it is create recorder case and recording is True
    #   OR
    # If it is update recorder case and recording status is changed
    if (not before and recording) or (before and recording != before["recording"]):
        if recording:
            recording_status_ret = (
                await hub.exec.boto3.client.config.start_configuration_recorder(
                    ctx=ctx, ConfigurationRecorderName=name
                )
            )
        else:
            recording_status_ret = (
                await hub.exec.boto3.client.config.stop_configuration_recorder(
                    ctx=ctx, ConfigurationRecorderName=name
                )
            )
        if not recording_status_ret["result"]:
            result["comment"] = recording_status_ret["comment"]
            result["result"] = False
            return result

        after_recording_status_ret = (
            await hub.exec.boto3.client.config.describe_configuration_recorder_status(
                ctx,
                ConfigurationRecorderNames=[name],
            )
        )
        if not after_recording_status_ret["result"]:
            result["comment"] = after_recording_status_ret["comment"]
            result["result"] = False
            return result
        result["new_state"]["recording"] = after_recording_status_ret["ret"][
            "ConfigurationRecordersStatus"
        ][0]["recording"]
        if only_recording_status_changed:
            result["comment"] = hub.tool.aws.comment_utils.update_comment(
                resource_type="aws.config.config_recorder", name=name
            )
    else:
        if not before:
            result["new_state"]["recording"] = False
        else:
            result["new_state"]["recording"] = before["recording"]
    return result


async def absent(hub, ctx, name: str, resource_id: str = None) -> Dict[str, Any]:
    """
    Deletes the configuration recorder. After the configuration recorder is deleted,
    Config will not record resource configuration changes until you create a new configuration recorder.

    Args:
        name(Text): The name of the recorder.
        resource_id(Text, optional): AWS Config configuration recorder Name. Idem automatically considers this resource being absent
         if this field is not specified.

    Returns:
          Dict[str, Any]

    Examples:
          .. code-block:: sls

            aws-config-recorder:
              aws.config.config_recorder.present:
                - name: 'config_recorder'
                - resource_id: 'config_recorder'

    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    before = None
    if resource_id:
        resource = await hub.exec.boto3.client.config.describe_configuration_recorders(
            ctx,
            ConfigurationRecorderNames=[name],
        )
        if not resource["result"]:
            if "NoSuchConfigurationRecorderException" in str(resource["comment"]):
                result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
                    resource_type="aws.config.config_recorder", name=name
                )
                return result
            result["comment"] = resource["comment"]
            result["result"] = False
            return result
        else:
            before = resource["ret"]["ConfigurationRecorders"][0]

    if before:
        before_recording = (
            await hub.exec.boto3.client.config.describe_configuration_recorder_status(
                ctx,
                ConfigurationRecorderNames=[name],
            )
        )
        if not before_recording["result"]:
            result["comment"] = before_recording["comment"]
            result["result"] = False
            return result

        translated_resource = await hub.tool.aws.config.conversion_utils.convert_raw_config_recorder_to_present(
            ctx,
            raw_resource=before,
            recording=before_recording["ret"]["ConfigurationRecordersStatus"][0][
                "recording"
            ],
        )
        result["old_state"] = translated_resource
    else:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.config.config_recorder", name=name
        )
        return result

    if ctx.get("test", False):
        result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
            resource_type="aws.config.config_recorder", name=name
        )
        return result
    else:
        delete_ret = await hub.exec.boto3.client.config.delete_configuration_recorder(
            ctx, ConfigurationRecorderName=name
        )
        result["result"] = delete_ret["result"]
        if not result["result"]:
            result["comment"] = delete_ret["comment"]
            return result
        result["comment"] = hub.tool.aws.comment_utils.delete_comment(
            resource_type="aws.config.config_recorder", name=name
        )
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    """
    Describe the resource in a way that can be recreated/managed with the corresponding "present" function

    Return details about your config recorder.

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: bash

            $ idem describe aws.config.config_recorder
    """
    result = {}
    ret = await hub.exec.boto3.client.config.describe_configuration_recorders(ctx)
    if not ret["result"]:
        hub.log.debug(f"Could not describe configuration recorder {ret['comment']}")
        return {}

    for resource in ret["ret"]["ConfigurationRecorders"]:
        name = resource["name"]
        recording_status = (
            await hub.exec.boto3.client.config.describe_configuration_recorder_status(
                ctx,
                ConfigurationRecorderNames=[name],
            )
        )
        if not recording_status["result"]:
            hub.log.warning(
                f"Could not fetch configuration recorder '{name}' status with error {recording_status['comment']}"
            )
            continue

        recorder_name = f"{name}-config-recorder"
        translated_resource = await hub.tool.aws.config.conversion_utils.convert_raw_config_recorder_to_present(
            ctx,
            raw_resource=resource,
            recording=recording_status["ret"]["ConfigurationRecordersStatus"][0][
                "recording"
            ],
        )
        result[recorder_name] = {
            "aws.config.config_recorder.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in translated_resource.items()
            ]
        }
    return result



"""
Contains functions that are useful for describing instances in a consistent manner
"""
from typing import Any
from typing import Dict


async def convert_to_present(
    hub, describe_spot_instance_requests: Dict[str, Any]
) -> Dict[str, Any]:
    """
    Convert instances from ec2.describe_spot_instance_requests() to aws.ec2.spot_instance_requests.present states

    This is the preferred "meta" function for collecting information about multiple instances
    """
    result = {}
    for spot_instance_request in describe_spot_instance_requests[
        "SpotInstanceRequests"
    ]:
        resource_id = spot_instance_request["SpotInstanceRequestId"]
        result[resource_id] = dict(
            name=resource_id,
            resource_id=resource_id,
            launch_specification=spot_instance_request.get("LaunchSpecification"),
            spot_price=spot_instance_request.get("SpotPrice"),
            type=spot_instance_request.get("Type"),
            availability_zone_group=spot_instance_request.get(
                "LaunchedAvailabilityZone"
            ),
            valid_until=spot_instance_request.get("ValidUntil"),
            instance_interruption_behavior=spot_instance_request.get(
                "InstanceInterruptionBehavior"
            ),
            # These don't appear in describe spot_instance_requests; can only be specified on creation
            instance_count=None,
            launch_group=None,
            valid_from=None,
        )
        if "Tags" in spot_instance_request:
            result[resource_id][
                "tags"
            ] = hub.tool.aws.tag_utils.convert_tag_list_to_dict(
                spot_instance_request.get("Tags")
            )
    return result


async def get(hub, ctx, *, resource_id: str) -> Dict[str, Any]:
    """
    Get the state of a single spot_instance_request from the conversion to present

    This is the preferred "meta" function for collecting information about a single instance
    """
    ret = await hub.exec.boto3.client.ec2.describe_spot_instance_requests(
        ctx, SpotInstanceRequestIds=[resource_id]
    )

    if not ret:
        return {}

    present_states = await hub.tool.aws.ec2.spot_instance_request.convert_to_present(
        ret.ret
    )
    if not present_states:
        return {}

    # There will only be one result from "convert_to_present", return it as a plain dictionary
    state = next(iter((present_states).values()))
    return state


def test(hub, **kwargs) -> Dict[str, Any]:
    """
    Compute the state based on the parameters passed to an instance.present function for ctx.test
    """
    result = {}
    for k, v in kwargs.items():
        # Ignore kwargs that were None
        if v is None:
            continue
        result[k] = v
    return result



import copy
from collections import OrderedDict
from typing import Any
from typing import Dict
from typing import List


async def update_web_acl(
    hub,
    ctx,
    name: str,
    raw_resource: Dict[str, Any],
    resource_parameters: Dict[str, None],
    scope: str,
    resource_id: str,
    lock_token: str,
):
    """

    Args:
        hub:
        ctx:
        name: Name of resource going to update.
        raw_resource: Old state of resource or existing resource details.
        resource_parameters: Parameters from sls file
        scope: Specifies whether this is for an Amazon CloudFront distribution or for a regional application.
        resource_id: The unique identifier for the web ACL.
        lock_token: A token used for optimistic locking.

    Returns:
        {"result": True|False, "comment": ("A tuple",), "ret": {}}
    """

    parameters = OrderedDict(
        {
            "Name": "name",
            "DefaultAction": "default_action",
            "VisibilityConfig": "visibility_config",
            "Description": "description",
            "Rules": "rules",
            "CustomResponseBodies": "custom_response_bodies",
            "CaptchaConfig": "captcha_config",
        }
    )
    parameters_to_update = {}
    result = dict(comment=(), result=True, ret=None)
    resource_parameters.pop("Tags", None)

    for key, value in resource_parameters.items():
        if value is None or value == raw_resource[key]:
            continue
        parameters_to_update[key] = resource_parameters[key]

    if parameters_to_update:
        if not ctx.get("test", False):
            update_ret = await hub.exec.boto3.client.wafv2.update_web_acl(
                ctx,
                Scope=scope,
                Id=resource_id,
                LockToken=lock_token,
                **resource_parameters,
            )
            if not update_ret["result"]:
                result["comment"] = update_ret["comment"]
                result["result"] = False
                return result
            result["comment"] = result["comment"] + (f"Updated '{name}'",)

        result["ret"] = {}
        for parameter_raw, parameter_present in parameters.items():
            if parameter_raw in parameters_to_update:
                result["ret"][parameter_present] = parameters_to_update[parameter_raw]

    return result


async def update_web_acl_tags(
    hub,
    ctx,
    web_acl_arn,
    old_tags: List[Dict[str, Any]] or Dict[str, Any],
    new_tags: List[Dict[str, Any]] or Dict[str, Any],
):
    """
    Update tags of Web ACL resources

    Args:
        hub:
        ctx:
        web_acl_arn: aws resource arn
        old_tags: list of old tags in the format of [{"Key": tag-key, "Value": tag-value}] or dict in the format of
         {tag-key: tag-value}
        new_tags: list of new tags in the format of [{"Key": tag-key, "Value": tag-value}] or dict in the format of
         {tag-key: tag-value}. If this value is None, the function will do no operation on tags.


    Returns:
        {"result": True|False, "comment": "A message", "ret": dict of updated tags}

    """

    tags_to_add = {}
    tags_to_remove = {}
    if isinstance(old_tags, List):
        old_tags = hub.tool.aws.tag_utils.convert_tag_list_to_dict(old_tags)
    if isinstance(new_tags, List):
        new_tags = hub.tool.aws.tag_utils.convert_tag_list_to_dict(new_tags)
    if new_tags is not None:
        tags_to_remove, tags_to_add = hub.tool.aws.tag_utils.diff_tags_dict(
            old_tags=old_tags, new_tags=new_tags
        )
    result = dict(comment=(), result=True, ret={})
    if (not tags_to_remove) and (not tags_to_add):
        result["ret"] = copy.deepcopy(old_tags if old_tags else {})
        return result

    if tags_to_remove:
        if not ctx.get("test", False):
            delete_ret = await hub.exec.boto3.client.wafv2.untag_resource(
                ctx, ResourceARN=web_acl_arn, TagKeys=list(tags_to_remove)
            )
            if not delete_ret["result"]:
                result["comment"] = delete_ret["comment"]
                result["result"] = False
                return result

    if tags_to_add:
        if not ctx.get("test", False):
            add_ret = await hub.exec.boto3.client.wafv2.tag_resource(
                ctx,
                ResourceARN=web_acl_arn,
                Tags=hub.tool.aws.tag_utils.convert_tag_dict_to_list(tags_to_add),
            )
            if not add_ret["result"]:
                result["comment"] = add_ret["comment"]
                result["result"] = False
                return result

    result["ret"] = new_tags
    result["comment"] = hub.tool.aws.comment_utils.update_tags_comment(
        tags_to_remove=tags_to_remove, tags_to_add=tags_to_add
    )
    return result



from typing import Dict
from typing import List

__func_alias__ = {"list_": "list"}


async def get(
    hub,
    ctx,
    name,
    resource_id: str = None,
    default: bool = None,
    filters: List = None,
) -> Dict:
    """
    Use an un-managed VPC as a data-source. Supply one of the inputs as the filter.

    Args:
        name(string): The name of the Idem state.
        resource_id(string, optional): AWS VPC id to identify the resource.
        default(bool, optional): Indicate whether the VPC is the default VPC.
        filters(list, optional): One or more filters: for example, tag :<key>, tag-key. A complete list of filters can be found at
         https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/ec2.html#EC2.Client.describe_vpcs

    """
    result = dict(comment=[], ret=None, result=True)
    ret = await hub.tool.aws.ec2.vpc.search_raw(
        ctx=ctx,
        name=name,
        resource_id=resource_id,
        default=default,
        filters=filters,
    )
    if not ret["result"]:
        if "InvalidVpcID.NotFound" in str(ret["comment"]):
            result["comment"].append(
                hub.tool.aws.comment_utils.get_empty_comment(
                    resource_type="aws.ec2.vpc", name=name
                )
            )
            result["comment"] += list(ret["comment"])
            return result
        result["comment"] += list(ret["comment"])
        result["result"] = False
        return result
    if not ret["ret"]["Vpcs"]:
        result["comment"].append(
            hub.tool.aws.comment_utils.get_empty_comment(
                resource_type="aws.ec2.vpc", name=name
            )
        )
        return result

    resource = ret["ret"]["Vpcs"][0]
    if len(ret["ret"]["Vpcs"]) > 1:
        result["comment"].append(
            f"More than one aws.ec2.vpc resource was found. Use resource {resource.get('VpcId')}"
        )
    result["ret"] = await hub.tool.aws.ec2.conversion_utils.convert_raw_vpc_to_present(
        ctx, raw_resource=resource, idem_resource_name=name
    )
    return result


async def list_(hub, ctx, name, filters: List = None, default: bool = None) -> Dict:
    """
    Use an un-managed VPC as a data-source. Supply one of the inputs as the filter.

    Args:
        name(string): The name of the Idem state.
        default(bool, optional): Indicate whether the VPC is the default VPC.
        filters(list, optional): One or more filters: for example, tag :<key>, tag-key. A complete list of filters can be found at
         https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/ec2.html#EC2.Client.describe_vpcs

    """
    result = dict(comment=[], ret=[], result=True)
    ret = await hub.tool.aws.ec2.vpc.search_raw(
        ctx=ctx,
        name=name,
        default=default,
        filters=filters,
    )
    if not ret["result"]:
        result["comment"] += list(ret["comment"])
        result["result"] = False
        return result
    if not ret["ret"]["Vpcs"]:
        result["comment"].append(
            hub.tool.aws.comment_utils.list_empty_comment(
                resource_type="aws.ec2.vpc", name=name
            )
        )
        return result
    for vpc in ret["ret"]["Vpcs"]:
        vpc_id = vpc.get("VpcId")
        result["ret"].append(
            await hub.tool.aws.ec2.conversion_utils.convert_raw_vpc_to_present(
                ctx, raw_resource=vpc, idem_resource_name=vpc_id
            )
        )
    return result



"""
Autogenerated state module using `pop-create-idem <https://gitlab.com/saltstack/pop/pop-create-idem>`__

hub.exec.boto3.client.ce.create_cost_category_definition
hub.exec.boto3.client.ce.delete_cost_category_definition
hub.exec.boto3.client.ce.describe_cost_category_definition
hub.exec.boto3.client.ce.update_cost_category_definition
"""
import copy
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    rules: List[Dict[str, Any]],
    cost_category_name: str,
    rule_version: str,
    resource_id: str = None,
    split_charge_rules: List[Dict[str, Any]] = None,
    default_value: str = None,
    tags: Dict[str, Any] = None,
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Creates a new Cost Category with the requested name and rules.

    Args:
        name(Text): The unique name of the Cost Category.
        rules(List[Dict[str, Any]]): The Cost Category rules used to categorize costs.
        cost_category_name(Text): Name of Cost Category.
        rule_version(Text): The rule schema version in this particular Cost Category.
        resource_id(Text, Optional): Cost Category ARN to identify the resource.
        split_charge_rules(List[Dict[str, Any]], Optional): The split charge rules used to allocate your charges between your Cost Category values.
        default_value(Text, Optional): The default value for the cost category.
        tags (List[Dict[str, Any]], Optional): Each tag consists of a key and a value, and each key must be unique for the resource.

    Request Syntax:
        [monitor-resource-id]:
          aws.costexplorer.cost_category.present:
          - resource_id: 'string'
          - tags: 'dict'
          - cost_category_name: 'string'
          - effective_start: 'string'
          - rules:
                - Rule:
                    Dimensions:
                      Key: 'string'
                      MatchOptions:
                      - EQUALS
                      Values:
                      - 'string'
                  Type: 'string'
                  Value: 'string'
          - split_charge_rules:
                - Method: 'string'
                  Parameters:
                  - Type: 'string'
                    Values:
                    - 'string'
                    - 'string'
                  Source: 'string'
                  Targets:
                  - 'string'
              - processing_status:
                - Component: 'string'
                  Status: 'string'
              - default_value: 'string'

    Returns:
        Dict[str, str]

    Examples:

        .. code-block:: sls

            arn:aws:ce::1234567891012:costcategory/4e9662f0-5533-4fdf-8224-9bb9f82d0a39:
              aws.costexplorer.cost_category.present:
              - resource_id: arn:aws:ce::1234567891012:costcategory/4e9662f0-5533-4fdf-8224-9bb9f82d0a39
              - tags:
                  name: test_category
              - cost_category_name: test_category
              - effective_start: '2022-06-01T00:00:00Z'
              - rule_version: CostCategoryExpression.v1
              - rules:
                - Rule:
                    Dimensions:
                      Key: LINKED_ACCOUNT
                      MatchOptions:
                      - EQUALS
                      Values:
                      - '107488843946'
                  Type: REGULAR
                  Value: Alpha
                - Rule:
                    Dimensions:
                      Key: LINKED_ACCOUNT
                      MatchOptions:
                      - EQUALS
                      Values:
                      - 011922870716
                  Type: REGULAR
                  Value: Beta
                - Rule:
                    Dimensions:
                      Key: LINKED_ACCOUNT
                      MatchOptions:
                      - EQUALS
                      Values:
                      - '332986752459'
                  Type: REGULAR
                  Value: Gamma
              - split_charge_rules:
                - Method: FIXED
                  Parameters:
                  - Type: ALLOCATION_PERCENTAGES
                    Values:
                    - '40'
                    - '60'
                  Source: Alpha
                  Targets:
                  - Beta
                  - Gamma
              - processing_status:
                - Component: COST_EXPLORER
                  Status: PROCESSING
              - default_value: Other

    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    before = None
    resource_updated = False

    if resource_id:
        before = await hub.exec.boto3.client.ce.describe_cost_category_definition(
            ctx, CostCategoryArn=resource_id
        )
        if not before["result"]:
            result["result"] = False
            result["comment"] = before["comment"]
            return result

    if before:
        result[
            "old_state"
        ] = await hub.tool.aws.costexplorer.conversion_utils.convert_raw_cost_category_to_present(
            ctx,
            raw_resource=before["ret"]["CostCategory"],
            idem_resource_name=name,
        )
        plan_state = copy.deepcopy(result["old_state"])

        update_ret = await hub.exec.aws.costexplorer.cost_category.update_cost_category_definition(
            ctx,
            before=before["ret"]["CostCategory"],
            rules=rules,
            default_value=default_value,
            split_charge_rules=split_charge_rules,
        )
        result["comment"] = result["comment"] + update_ret["comment"]
        result["result"] = update_ret["result"]
        resource_updated = resource_updated or bool(update_ret["ret"])
        if update_ret["ret"] and ctx.get("test", False):
            plan_state = update_ret["ret"]

        if tags is not None and tags != result["old_state"].get("tags"):
            # Update tags
            update_ret = await hub.exec.aws.costexplorer.cost_category.update_tags(
                ctx,
                resource_id=resource_id,
                old_tags=hub.tool.aws.tag_utils.convert_tag_dict_to_list(
                    result["old_state"].get("tags")
                ),
                new_tags=hub.tool.aws.tag_utils.convert_tag_dict_to_list(tags),
            )
            result["result"] = result["result"] and update_ret["result"]
            result["comment"] = result["comment"] + update_ret["comment"]
            resource_updated = resource_updated or bool(update_ret["ret"])
            if ctx.get("test", False) and update_ret["ret"] is not None:
                plan_state["tags"] = hub.tool.aws.tag_utils.convert_tag_list_to_dict(
                    update_ret["ret"].get("tags")
                )

        if resource_updated:
            if ctx.get("test", False):
                result["comment"] += hub.tool.aws.comment_utils.would_update_comment(
                    resource_type="aws.costexplorer.cost_category", name=name
                )
            else:
                result["comment"] += hub.tool.aws.comment_utils.update_comment(
                    resource_type="aws.costexplorer.cost_category", name=name
                )
    else:
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "cost_category_name": cost_category_name,
                    "rules": rules,
                    "split_charge_rules": split_charge_rules,
                    "default_value": default_value,
                    "tags": tags,
                    "rule_version": rule_version,
                },
            )
            result["comment"] = hub.tool.aws.comment_utils.would_create_comment(
                resource_type="aws.costexplorer.cost_category", name=name
            )
            return result

        ret = await hub.exec.boto3.client.ce.create_cost_category_definition(
            ctx,
            Name=cost_category_name,
            RuleVersion=rule_version,
            Rules=rules,
            DefaultValue=default_value,
            SplitChargeRules=split_charge_rules,
            ResourceTags=hub.tool.aws.tag_utils.convert_tag_dict_to_list(tags),
        )
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            return result
        result["comment"] = hub.tool.aws.comment_utils.create_comment(
            resource_type="aws.costexplorer.cost_category", name=name
        )
        resource_id = ret["ret"]["CostCategoryArn"]

    try:
        if ctx.get("test", False):
            result["new_state"] = plan_state
        elif (not before) or resource_updated:
            after = await hub.exec.boto3.client.ce.describe_cost_category_definition(
                ctx, CostCategoryArn=resource_id
            )
            if not after["result"]:
                result["result"] = False
                result["comment"] = after["comment"]
                return result
            result[
                "new_state"
            ] = await hub.tool.aws.costexplorer.conversion_utils.convert_raw_cost_category_to_present(
                ctx,
                raw_resource=after["ret"]["CostCategory"],
                idem_resource_name=name,
            )
        else:
            result["new_state"] = copy.deepcopy(result["old_state"])
    except Exception as e:
        result["comment"] = result["comment"] + (str(e),)
        result["result"] = False
    return result


async def absent(hub, ctx, name: str, resource_id: str = None) -> Dict[str, Any]:
    """
    **Autogenerated function**

    Deletes a cost category.

    Args:
        name(Text): An Idem name of the resource.
        resource_id(Text, Optional): Cost Category ARN to identify the resource. Idem automatically considers this resource being absent if this field is not specified.

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            resource_is_absent:
              aws.costexplorer.cost_category.absent:
                - name: value
                - resource_id: value
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)

    if not resource_id:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.costexplorer.cost_category", name=name
        )
        return result

    before = await hub.exec.boto3.client.ce.describe_cost_category_definition(
        ctx, CostCategoryArn=resource_id
    )

    if not before:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.costexplorer.cost_category", name=name
        )
    elif ctx.get("test", False):
        result[
            "old_state"
        ] = await hub.tool.aws.costexplorer.conversion_utils.convert_raw_cost_category_to_present(
            ctx, raw_resource=before["ret"]["CostCategory"], idem_resource_name=name
        )
        result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
            resource_type="aws.costexplorer.cost_category", name=name
        )
        return result
    else:
        result[
            "old_state"
        ] = await hub.tool.aws.costexplorer.conversion_utils.convert_raw_cost_category_to_present(
            ctx, raw_resource=before["ret"]["CostCategory"], idem_resource_name=name
        )
        ret = await hub.exec.boto3.client.ce.delete_cost_category_definition(
            ctx, CostCategoryArn=resource_id
        )
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            result["result"] = False
            return result
        result["comment"] = hub.tool.aws.comment_utils.delete_comment(
            resource_type="aws.costexplorer.cost_category", name=name
        )

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""
    **Autogenerated function**

    Describe the resource in a way that can be recreated/managed with the corresponding "present" function


    Returns a list of aws cost categories.


    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: bash

            $ idem describe aws.costexplorer.cost_category
    """
    result = {}
    ret = await hub.exec.boto3.client.ce.list_cost_category_definitions(ctx)

    if not ret["result"]:
        hub.log.debug(f"Could not list Cost Category Definitions {ret['comment']}")
        return {}

    for costcategory in ret["ret"]["CostCategoryReferences"]:
        resource_id = costcategory.get("CostCategoryArn")
        ret = await hub.exec.boto3.client.ce.describe_cost_category_definition(
            ctx, CostCategoryArn=resource_id
        )

        if not ret["result"]:
            hub.log.debug(
                f"Could not describe Cost Category Definitions {ret['comment']}"
            )
            return {}

        resource_translated = await hub.tool.aws.costexplorer.conversion_utils.convert_raw_cost_category_to_present(
            ctx,
            raw_resource=ret["ret"].get("CostCategory"),
            idem_resource_name=resource_id,
        )
        result[resource_id] = {
            "aws.costexplorer.cost_category.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource_translated.items()
            ]
        }
    return result
